/* information about connections between hosts and clients
 * Copyright (C) 1998, 1999  D. Hugh Redelmeier.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: connections.c,v 1.79 2000/11/29 05:53:15 dhr Exp $
 */

#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <freeswan.h>

#include "constants.h"
#include "defs.h"
#include "id.h"
#include "connections.h"	/* needs id.h */
#include "packet.h"
#include "demux.h"	/* needs packet.h */
#include "state.h"
#include "timer.h"
#include "ipsec_doi.h"	/* needs demux.h and state.h */
#include "server.h"
#include "kernel.h"	/* for no_klips */
#include "log.h"
#include "preshared.h"
#include "dnskey.h"	/* needs preshared.h */
#include "whack.h"

static struct connection *connections = NULL;

/* host_pair: a nexus of information about a pair of hosts.
 * A host is an IP address, UDP port pair.  This is a debatable choice:
 * - should port be considered (no choice of port in standard)?
 * - should ID be considered (hard because not always known)?
 * - should IP address matter on our end (we don't know our end)?
 * Since connections are not all oriented, the host identifiers
 * (ip_address values) are placed in a canonical order.
 */

struct host_pair {
    struct {
	ip_address addr;
	u_int16_t port;	/* host order */
    } a, b;
    bool initial_connection_sent;
    struct connection *connections;	/* connections with this pair */
    struct host_pair *next;
};

static struct host_pair *host_pairs = NULL;

/* check to see that Ids of peers match */
bool
same_peer_ids(const struct connection *c, const struct connection *d
, const struct id *his_id)
{
    return same_id(&c->this.id, &d->this.id)
	&& same_id(his_id == NULL? &c->that.id : his_id, &d->that.id);
}

/* Arbitrary canonicalization of host pair.
 * Because it is arbitrary, we don't need to compare in host order.
 */
static void
canonical_host_pair(struct host_pair *hp
, const ip_address *xaddr, u_int16_t xport
, const ip_address *yaddr, u_int16_t yport)
{
    int ac;

    passert(sameaddrtype(xaddr, yaddr));
    ac = addrcmp(xaddr, yaddr);

    if (ac < 0
    || (ac == 0 && xport <= yport))
    {
	hp->a.addr = *xaddr;
	hp->a.port = xport;
	hp->b.addr = *yaddr;
	hp->b.port = yport;
    }
    else
    {
	hp->a.addr = *yaddr;
	hp->a.port = yport;
	hp->b.addr = *xaddr;
	hp->b.port = xport;
    }

    return;
}

static struct host_pair *
find_host_pair(const ip_address *xaddr, u_int16_t xport
, const ip_address *yaddr, u_int16_t yport)
{
    struct host_pair hp;
    struct host_pair *p, *prev;
    ip_address any;

    if (yaddr == NULL)
    {
	(void) anyaddr(addrtypeof(xaddr), &any);
	yaddr = &any;
    }

    canonical_host_pair(&hp, xaddr, xport, yaddr, yport);

    for (prev = NULL, p = host_pairs; p != NULL; prev = p, p = p->next)
    {
	if (sameaddr(&p->a.addr, &hp.a.addr) && p->a.port == hp.a.port
	&& sameaddr(&p->b.addr, &hp.b.addr) && p->b.port == hp.b.port)
	{
	    if (prev != NULL)
	    {
		prev->next = p->next;	/* remove p from list */
		p->next = host_pairs;	/* and stick it on front */
		host_pairs = p;
	    }
	    break;
	}
    }
    return p;
}

static struct connection *
find_host_pair_connections(const ip_address *xaddr, u_int16_t xport
, const ip_address *yaddr, u_int16_t yport)
{
    struct host_pair *hp = find_host_pair(xaddr, xport, yaddr, yport);

    return hp == NULL? NULL : hp->connections;
}

static void
connect_to_host_pair(struct connection *c)
{
    struct host_pair *hp = find_host_pair(&c->this.host_addr, c->this.host_port
	, &c->that.host_addr, c->that.host_port);

    if (hp == NULL)
    {
	hp = alloc_thing(struct host_pair, "host_pair");
	canonical_host_pair(hp, &c->this.host_addr, c->this.host_port
	    , &c->that.host_addr, c->that.host_port);
	hp->initial_connection_sent = FALSE;
	hp->connections = NULL;
	hp->next = host_pairs;
	host_pairs = hp;
    }
    c->host_pair = hp;
    c->hp_next = hp->connections;
    hp->connections = c;
}

/* figure out if Opportunism is possible: are there any connections
 * with wildcarded peer AND peer's client?
 * Result is remembered until an addition or deletion makes it suspect.
 */
static bool
    definitely_some_oppo = FALSE,
    possibly_some_oppo = FALSE;

bool
opportunism_possible(void)
{
    if (definitely_some_oppo != possibly_some_oppo)
    {
	/* hard work: we've got to look */
	struct iface *p;
	/* for each of our addresses... */
	for (p = interfaces; p != NULL; p = p->next)
	{
	    /* go through those connections with our address and NO_IP as hosts
	     * We cannot know what port the peer would use, so we assume
	     * that it is pluto_port (makes debugging easier).
	     */
	    struct connection *c = find_host_pair_connections(&p->addr
		, pluto_port, (ip_address *)NULL, pluto_port);

	    for (; c != NULL; c = c->next)
	    {
		if (c->rw_state == rwcs_permanent && HasWildcardClient(c))
		{
		    passert(HasWildcardIP(*c));
		    return definitely_some_oppo = possibly_some_oppo = TRUE;
		}
	    }
	}
	definitely_some_oppo = possibly_some_oppo = FALSE;
    }
    return definitely_some_oppo;
}

/* find a connection by name.
 * move the winner (if any) to the front.
 * If none is found, and wfd is not NULL_FD, a diagnostic is logged to whack.
 */
struct connection *
con_by_name(const char *nm, bool strict)
{
    struct connection *p, *prev;

    for (prev = NULL, p = connections; ; prev = p, p = p->next)
    {
	if (p == NULL)
	{
	    if (strict)
		whack_log(RC_UNKNOWN_NAME, "no connection with that name");
	    break;
	}
	if (streq(p->name, nm))
	{
	    if (prev != NULL)
	    {
		prev->next = p->next;	/* remove p from list */
		p->next = connections;	/* and stick it on front */
		connections = p;
	    }
	    break;
	}
    }
    return p;
}

void
unorient_connection(struct connection *c)
{
    if (c->rw_state == rwcs_instance)
    {
	/* This does everything we need.
	 * Note that we will be called by delete_connection,
	 * but rw_state will be rwcs_going_away.
	 */
	delete_connection(c);
    }
    else
    {
	delete_states_by_connection(c);
	unroute_connection(c);
	c->interface = NULL;
    }
}

/* Delete a connection */

#define list_rm(etype, enext, e, ehead) { \
	etype **ep; \
	for (ep = &(ehead); *ep != (e); ep = &(*ep)->enext) \
	    passert(*ep != NULL);    /* we must not come up empty-handed */ \
	*ep = (e)->enext; \
    }


void
delete_connection(struct connection *c)
{
    struct connection *old_cur_connection
	= cur_connection == c? NULL : cur_connection;
#ifdef DEBUG
    unsigned int old_cur_debugging = cur_debugging;
#endif

    SET_CUR_CONNECTION(c);

    /* Must be careful to avoid circularity:
     * we mark c as going away so it won't get deleted recursively.
     */
    passert(c->rw_state != rwcs_going_away);
    if (c->rw_state == rwcs_instance)
    {
	log("deleting connection \"%s\" with Road Warrior %s"
	    , c->name, ip_str(&c->that.host_addr));
	c->rw_state = rwcs_going_away;
    }
    else
    {
	definitely_some_oppo = FALSE;	/* could be more careful */
	log("deleting connection");
    }
    unorient_connection(c);	/* won't delete c */

    /* find and delete c from connections list */
    list_rm(struct connection, next, c, connections);
    cur_connection = old_cur_connection;

    /* find and delete c from the host pair list */
    {
	struct host_pair *hp = c->host_pair;

	list_rm(struct connection, hp_next, c, hp->connections);
	c->host_pair = NULL;	/* redundant, but safe */

	/* if there are no more connections with this host_pair
	 * and we haven't even made an initial contact, let's delete
	 * this guy in case we were created by an attempted DOS attack.
	 */
	if (hp->connections == NULL && !hp->initial_connection_sent)
	{
	    list_rm(struct host_pair, next, hp, host_pairs);
	    pfree(hp);
	}
    }

#ifdef DEBUG
    cur_debugging = old_cur_debugging;
#endif
    pfreeany(c->name);
    free_id_content(&c->this.id);
    pfreeany(c->this.updown);
    free_id_content(&c->that.id);
    pfreeany(c->that.updown);
    pfree(c);
}

void
delete_every_connection(void)
{
    while (connections != NULL)
	delete_connection(connections);
}

void
release_interface(struct iface *i)
{
    struct connection *c, *n;

    passert(i != NULL);
    for (c = connections; c != NULL; c = n)
    {
	n = c->next;
	if (c->interface == i)
	    unorient_connection(c);	/* may delete c */
    }
}

static err_t
default_end(struct end *e, ip_address *dflt_nexthop)
{
    err_t ugh = NULL;
    const struct af_info *afi = aftoinfo(addrtypeof(&e->host_addr));

    if (afi == NULL)
	return "unknown address family in default_end";

    /* default ID to IP (but only if not NO_IP -- WildCard) */
    if (e->id.kind == ID_NONE && !isanyaddr(&e->host_addr))
    {
	e->id.kind = afi->id_addr;
	e->id.ip_addr = e->host_addr;
    }

    /* default nexthop to other side */
    if (isanyaddr(&e->host_nexthop))
	e->host_nexthop = *dflt_nexthop;

    /* default client to subnet containing only self
     * XXX This may mean that the client's address family doesn't match
     * tunnel_addr_family.
     */
    if (!e->has_client)
	ugh = initsubnet(&e->host_addr, afi->mask_cnt, '0', &e->client);

    return ugh;
}

/* format the topology of an end, leaving out defaults
 * Note: if that==NULL, skip nexthop
 */
size_t
format_end(char *buf, size_t buf_len
, const struct end *this, const struct end *that, bool is_left)
{
    char client[SUBNETTOT_BUF];
    const char *client_sep = "";
    char host[ADDRTOT_BUF];
    char host_port[sizeof(":65535")];
    char host_id[IDTOA_BUF];
    char hop[ADDRTOT_BUF];
    const char *hop_sep = "";
    ip_address net;

    client[0] = '\0';

    if (isanyaddr(&this->host_addr)
    && this->has_client
    && subnetishost(&this->client)
    && (networkof(&this->client, &net), isanyaddr(&net)))
    {
	/* %opportunistic subsumes client and host */
	snprintf(host, sizeof(host), "%%opportunistic");
    }
    else
    {
	/* [client===] */
	if (this->has_client)
	{
	    subnettot(&this->client, 0, client, sizeof(client));
	    client_sep = "===";
	}

	/* host */
	if (isanyaddr(&this->host_addr))
	    snprintf(host, sizeof(host), "%%any");
	else
	    addrtot(&this->host_addr, 0, host, sizeof(host));
    }
    host_port[0] = '\0';
    if (this->host_port != IKE_UDP_PORT)
	snprintf(host_port, sizeof(host_port), ":%u"
	    , this->host_port);

    /* id, if different from host */
    host_id[0] = '\0';
    if (!(this->id.kind == ID_NONE
    || ((this->id.kind == ID_IPV4_ADDR || this->id.kind == ID_IPV6_ADDR) && sameaddr(&this->id.ip_addr, &this->host_addr))))
    {
	int len = idtoa(&this->id, host_id+1, sizeof(host_id)-2);

	host_id[0] = '[';
	strcpy(&host_id[len < 0? (ptrdiff_t)sizeof(host_id)-2 : 1 + len], "]");
    }

    /* [---hop] */
    hop[0] = '\0';
    hop_sep = "";
    if (that != NULL && !sameaddr(&this->host_nexthop, &that->host_addr))
    {
	addrtot(&this->host_nexthop, 0, hop, sizeof(hop));
	hop_sep = "---";
    }

    if (is_left)
	snprintf(buf, buf_len, "%s%s%s%s%s%s%s"
	    , client, client_sep
	    , host, host_port, host_id
	    , hop_sep, hop);
    else
	snprintf(buf, buf_len, "%s%s%s%s%s%s%s"
	    , hop, hop_sep
	    , host, host_port, host_id
	    , client_sep, client);
    return strlen(buf);
}

static void
unshare_connection_strings(struct connection *c)
{
    c->name = clone_str(c->name, "connection name");

    unshare_id_content(&c->this.id);
    c->this.updown = clone_str(c->this.updown, "updown");
    unshare_id_content(&c->that.id);
    c->that.updown = clone_str(c->that.updown, "updown");
}

static void
extract_end(struct end *dst, const struct whack_end *src, const char *which)
{
    /* decode id, if any */
    if (src->id == NULL)
    {
	dst->id.kind = ID_NONE;
    }
    else
    {
	err_t ugh = atoid(src->id, &dst->id);

	if (ugh != NULL)
	{
	    loglog(RC_BADID, "bad %s --id: %s (ignored)", which, ugh);
	    dst->id = empty_id;	/* ignore bad one */
	}
    }

    /* the rest is simple copying of corresponding fields */
    dst->host_addr = src->host_addr;
    dst->host_nexthop = src->host_nexthop;
    dst->client = src->client;

    dst->has_client = src->has_client;
    dst->updown = src->updown;
    dst->host_port = src->host_port;
}

static bool
check_connection_end(const struct whack_end *this, const struct whack_end *that
, const struct whack_message *wm)
{
    if (wm->addr_family != addrtypeof(&this->host_addr)
    || wm->addr_family != addrtypeof(&this->host_nexthop)
    || (this->has_client? wm->tunnel_addr_family : wm->addr_family)
      != subnettypeof(&this->client)
    || subnettypeof(&this->client) != subnettypeof(&that->client))
    {
	/* this should have been diagnosed by whack, so we need not be clear */
	loglog(RC_CLASH, "address family inconsistency in connection");
	return FALSE;
    }

    if (isanyaddr(&that->host_addr))
    {
	/* other side is wildcard: we must check if other conditions met */
	if (isanyaddr(&this->host_addr))
	{
	    loglog(RC_ORIENT, "connection must specify host IP address for our side");
	    return FALSE;
	}
	else
	{
	    /* check that all RW IKE policies agree because we must implement
	     * them before the correct connection is known.
	     * We cannot enforce this for other non-RW connections because
	     * differentiation is possible when a command specifies which
	     * to initiate.
	     */
	    const struct connection *c = NULL;

	    c = find_host_pair_connections(&this->host_addr
		, this->host_port, (const ip_address *)NULL, that->host_port);

	    for (; c != NULL; c = c->hp_next)
	    {
		if ((c->policy ^ wm->policy) & (POLICY_PSK | POLICY_RSASIG))
		{
		    loglog(RC_CLASH
			, "authentication method disagrees with \"%s\", which is also for an unspecified peer"
			, c->name);
		    return FALSE;
		}
	    }
	}
    }
    return TRUE;	/* happy */
}

void
add_connection(const struct whack_message *wm)
{
    if (con_by_name(wm->name, FALSE) != NULL)
    {
	loglog(RC_DUPNAME, "attempt to redefine connection \"%s\"", wm->name);
    }
    else if (check_connection_end(&wm->right, &wm->left, wm)
    && check_connection_end(&wm->left, &wm->right, wm))
    {
	struct connection *c = alloc_thing(struct connection, "struct connection");

	c->name = wm->name;

	c->policy = wm->policy;

	c->sa_ike_life_seconds = wm->sa_ike_life_seconds;
	c->sa_ipsec_life_seconds = wm->sa_ipsec_life_seconds;
	c->sa_rekey_margin = wm->sa_rekey_margin;
	c->sa_rekey_fuzz = wm->sa_rekey_fuzz;
	c->sa_keying_tries = wm->sa_keying_tries;

	c->addr_family = wm->addr_family;
	c->tunnel_addr_family = wm->tunnel_addr_family;

	extract_end(&c->this, &wm->left, "left");
	extract_end(&c->that, &wm->right, "right");

	default_end(&c->this, &c->that.host_addr);
	default_end(&c->that, &c->this.host_addr);

	/* force any wildcard host IP address to that end */
	if (isanyaddr(&wm->left.host_addr))
	{
	    struct end t = c->this;

	    c->this = c->that;
	    c->that = t;
	}

	/* set internal fields */
	c->next = connections;
	connections = c;
	c->interface = NULL;
	c->routed = FALSE;
	c->newest_isakmp_sa = SOS_NOBODY;
	c->newest_ipsec_sa = SOS_NOBODY;
	c->eroute_owner = SOS_NOBODY;

	c->rw_state = rwcs_permanent;

#ifdef DEBUG
	c->extra_debugging = wm->debugging;
#endif

	c->gw_info = NULL;

	unshare_connection_strings(c);

	connect_to_host_pair(c);

	possibly_some_oppo = TRUE;	/* could be more careful about this */

	/* log all about this connection */
	log("added connection description \"%s\"", c->name);
	DBG(DBG_CONTROL,
	    char lhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];
	    char rhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];

	    (void) format_end(lhs, sizeof(lhs), &c->this, &c->that, TRUE);
	    (void) format_end(rhs, sizeof(rhs), &c->that, &c->this, FALSE);

	    DBG_log("%s...%s", lhs, rhs);

	    /* Make sure that address families can be correctly inferred
	     * from printed ends.
	     */
	    passert(c->addr_family == addrtypeof(&c->this.host_addr)
		&& c->addr_family == addrtypeof(&c->this.host_nexthop)
		&& (c->this.has_client? c->tunnel_addr_family : c->addr_family)
		  == subnettypeof(&c->this.client)

		&& c->addr_family == addrtypeof(&c->that.host_addr)
		&& c->addr_family == addrtypeof(&c->that.host_nexthop)
		&& (c->that.has_client? c->tunnel_addr_family : c->addr_family)
		  == subnettypeof(&c->that.client));

	    DBG_log("ike_life: %lus; ipsec_life: %lus; rekey_margin: %lus;"
		" rekey_fuzz: %lu%%; keyingtries: %lu; policy: %s"
		, (unsigned long) c->sa_ike_life_seconds
		, (unsigned long) c->sa_ipsec_life_seconds
		, (unsigned long) c->sa_rekey_margin
		, (unsigned long) c->sa_rekey_fuzz
		, (unsigned long) c->sa_keying_tries
		, bitnamesof(sa_policy_bit_names, c->policy));
	);
    }
}

/* Common part of instantiating a Road Warrior or Opportunistic connection.
 * his_id can be used to carry over an ID discovered in Phase 1.
 * It must not disagree with the one in c, but if that is unspecified,
 * the new connection will use his_id.
 * If his_id is NULL, and c.that.id is uninstantiated (ID_NONE), the
 * new connection will continue to have an uninstantiated that.id.
 * Note: instantiation does not affect port numbers.
 */
static struct connection *
instantiate(const struct connection *c, const ip_address *him
, const struct id *his_id)
{
    struct connection *d = clone_thing(*c, "temporary connection");

    passert(c->rw_state == rwcs_permanent);
    if (his_id != NULL)
    {
	passert(d->that.id.kind == ID_NONE || same_id(&d->that.id, his_id));
	d->that.id = *his_id;
    }
    unshare_connection_strings(d);

    d->rw_state = rwcs_instance;

    passert(oriented(*d));
    passert(HasWildcardIP(*d));
    d->that.host_addr = *him;
    default_end(&d->that, &d->this.host_addr);

    /* We cannot guess what our next_hop should be, but if it was
     * explicitly specified as 0.0.0.0, we set it to be him.
     * (whack will not allow nexthop to be elided in RW case.)
     */
    default_end(&d->this, &d->that.host_addr);

    /* set internal fields */
    d->next = connections;
    connections = d;
    d->routed = FALSE;
    d->newest_isakmp_sa = SOS_NOBODY;
    d->newest_ipsec_sa = SOS_NOBODY;
    d->eroute_owner = SOS_NOBODY;

    connect_to_host_pair(d);

    return d;
}

struct connection *
rw_instantiate(const struct connection *c
, const ip_address *him
, const struct id *his_id)
{
    struct connection *d = instantiate(c, him, his_id);

    DBG(DBG_CONTROL
	, DBG_log("instantiated \"%s\" for %s" , d->name, ip_str(him)));
    return d;
}

struct connection *
oppo_instantiate(const struct connection *c
, const ip_address *him
, const struct id *his_id
, struct gw_info *gw
, const ip_subnet *our_client
, const ip_subnet *peer_client)
{
    struct connection *d = instantiate(c, him, his_id);

    /* fill in our client side */
    if (d->this.has_client)
    {
	/* there was a client in the abstract connection
	 * so we demand that the required client is within that subnet.
	 */
	passert(addrinsubnet(&d->this.host_addr, our_client));
	d->this.client = *our_client;
    }
    else
    {
	/* there was no client in the abstract connection
	 * so we demand that the required client be the host
	 */
	passert(subnetishost(our_client)
	    && addrinsubnet(&d->this.host_addr, our_client));
    }

    /* fill in peer's client side.
     * If the client is the peer, excise the client from the connection.
     */
    passert(HasWildcardClient(d));
    d->that.client = *peer_client;

    if (subnetishost(peer_client)
    && addrinsubnet(&d->that.host_addr, peer_client))
	d->that.has_client = FALSE;

    passert(d->gw_info == NULL);
    d->gw_info = gw;

    DBG(DBG_CONTROL,
	char lhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];
	char rhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];

	(void) format_end(lhs, sizeof(lhs), &d->this, &d->that, TRUE);
	(void) format_end(rhs, sizeof(rhs), &d->that, &d->this, FALSE);

	DBG_log("instantiated \"%s\": %s...%s", d->name, lhs, rhs);
    );
    return d;
}

/* Find an existing connection for an outgoing Opportunistic connection.
 * This is attempted before we bother with gateway discovery.
 *   + this is us (i.e. connection is oriented)
 *   + this subnet contains our_client (or we are our_client)
 *   + that subnet contains peer_client (or peer is peer_client)
 *   + don't care about peer (we don't know), but not NO_IP
 *   + don't care about Phase 1 IDs (we don't know)
 * Give preference to the first one that has an SA (either phase).
 */
struct connection *
find_opportunistic_connection(const ip_subnet *our_client, const ip_subnet *peer_client)
{
    struct connection *c = connections
	, *best = NULL;

    {
	ip_address network_of_our_client;
	ip_address network_of_peer_client;

	networkof(our_client, &network_of_our_client);
	networkof(peer_client, &network_of_peer_client);
	passert(!isanyaddr(&network_of_our_client)
	    && !isanyaddr(&network_of_peer_client));
    }

    for (c = connections; c != NULL; c = c->next)
    {
	if (oriented(*c) || orient(c, FALSE))
	{
	    if (addrinsubnet(&c->this.host_addr, our_client)
	    && addrinsubnet(&c->that.host_addr, peer_client))
	    {
		best = c;
		if (c->newest_isakmp_sa != SOS_NOBODY
		|| c->newest_ipsec_sa != SOS_NOBODY)
		    break;
	    }
	}
    }
    return best;
}

/* Find and instantiate a connection for an outgoing Opportunistic connection.
 * We've already discovered its gateway.
 *   + this is one of our interfaces
 *   + this subnet contains our_client (or we are our_client)
 *     (we will specialize the client)
 *   + that subnet is specified and is NO_IP/32 (we will fill in)
 *   + that peer is NO_IP (we will fill in)
 *   + don't care about Phase 1 IDs (probably should be default)
 * We could look for a connection that already had the desired peer
 * (rather than NO_IP) specified, but it doesn't seem worth the
 * bother.
 */
struct connection *
build_outgoing_opportunistic_connection(struct gw_info *gw
, const ip_subnet *our_client, const ip_subnet *peer_client)
{
    struct iface *p;

    {
	ip_address network_of_our_client;
	ip_address network_of_peer_client;

	networkof(our_client, &network_of_our_client);
	networkof(peer_client, &network_of_peer_client);
	passert(!isanyaddr(&network_of_our_client)
	    && !isanyaddr(&network_of_peer_client));
    }

    /* for each of our addresses... */
    for (p = interfaces; p != NULL; p = p->next)
    {
	/* go through those connections with our address and NO_IP as hosts
	 * We cannot know what port the peer would use, so we assume
	 * that it is pluto_port (makes debugging easier).
	 */
	struct connection *c = find_host_pair_connections(&p->addr
	    , pluto_port, (ip_address *)NULL, pluto_port);

	for (; c != NULL; c = c->next)
	{
	    if (!oriented(*c))
		(void)orient(c, TRUE);	/* needed by rw_instantiate */

	    if (addrinsubnet(&c->this.host_addr, our_client)
	    && HasWildcardClient(c))
	    {
		/* success: instantiate.
		 * We don't know his ID yet.
		 */
		return oppo_instantiate(c, &gw->gw, NULL, gw, our_client
		    , peer_client);
	    }
	}
    }
    return NULL;	/* no luck */
}


bool
orient(struct connection *c, bool must)
{
    if (!oriented(*c))
    {
	struct iface *p;

	/* Note: this loop does not stop when it finds a match:
	 * it continues checking to catch any ambiguity.
	 */
	for (p = interfaces; p != NULL; p = p->next)
	{
	    for (;;)
	    {
		/* check if this interface matches this end */
		if (sameaddr(&c->this.host_addr, &p->addr)
		&& (!no_klips || c->this.host_port == pluto_port))
		{
		    if (oriented(*c))
		    {
			if (c->interface == p)
			    loglog(RC_LOG_SERIOUS
				, "both sides of \"%s\" are our interfaces!"
				, c->name);
			else
			    loglog(RC_LOG_SERIOUS, "two interfaces match \"%s\"", c->name);
			c->interface = NULL;	/* withdraw orientation */
			return FALSE;
		    }
		    c->interface = p;
		}

		/* done with this interface if it doesn't match that end */
		if (!(sameaddr(&c->that.host_addr, &p->addr)
		&& (!no_klips || c->that.host_port == pluto_port)))
		    break;

		/* swap ends and try again */
		{
		    struct end t = c->this;

		    c->this = c->that;
		    c->that = t;
		}
	    }
	}
	if (!oriented(*c))
	{
	    if (must)
		loglog(RC_LOG_SERIOUS, "cannot find interface for connection \"%s\"", c->name);
	    return FALSE;
	}
    }
    return TRUE;
}

void
initiate_connection(const char *name, int whackfd)
{
    struct connection *c = con_by_name(name, TRUE);

    if (c != NULL)
    {
	SET_CUR_CONNECTION(c);
	if (!orient(c, TRUE))
	{
	    loglog(RC_ORIENT, "could not orient connection");
	}
	else if (HasWildcardIP(*c))
	{
	    loglog(RC_NOPEERIP, "cannot initiate connection without peer IP");
	}
	else
	{
	    /* We will only request an IPsec SA if policy isn't empty
	     * (ignoring Main Mode items).
	     * This is a fudge, but not yet important.
	     * If we are to proceed asynchronously, whackfd will be NULL_FD.
	     */
	    ipsecdoi_initiate(whackfd
		, c
		, (c->policy>>POLICY_IPSEC_SHIFT) != 0
		, c->policy
		, 1);
	    whackfd = NULL_FD;	/* protect from close */
	}
	UNSET_CUR_CONNECTION();
    }
    if (whackfd != NULL_FD)
	close(whackfd);
}

/* knowing clients (single IP addresses), try to build an SA */
void
initiate_opportunistic(const ip_subnet *our_client
, const ip_subnet *peer_client, int whackfd)
{
    struct connection *c;

    /* What connection shall we use?
     * First try for an existing one.
     */
    c = find_opportunistic_connection(our_client, peer_client);
    if (c != NULL)
    {
	/* We will only request an IPsec SA if policy isn't empty
	 * (ignoring Main Mode items).
	 * This is a fudge, but not yet important.
	 * If we are to proceed asynchronously, whackfd will be NULL_FD.
	 */
	ipsecdoi_initiate(whackfd
	    , c
	    , (c->policy>>POLICY_IPSEC_SHIFT) != 0
	    , c->policy
	    , 1);
	whackfd = NULL_FD;	/* protect from close */
    }
    else
    {
	/* try to find peer via DNS query */
	err_t ugh = discover_gateway(peer_client);

	if (ugh != NULL)
	{
	    loglog(RC_OPPOFAILURE, "failure discovering gateway: %s", ugh);
	}
	else
	{
	    /* note: new gw_info is at front of gateways list */
	    c = build_outgoing_opportunistic_connection(gateways, our_client
		, peer_client);
	    if (c == NULL)
	    {
		char ocb[SUBNETTOT_BUF]
		    , pcb[SUBNETTOT_BUF]
		    , pb[ADDRTOT_BUF];

		subnettot(our_client, 0, ocb, sizeof(ocb));
		subnettot(peer_client, 0, pcb, sizeof(ocb));
		addrtot(&gateways->gw, 0, pb, sizeof(pb));
		loglog(RC_OPPOFAILURE
		    , "no suitable connection for opportunism"
		      " between %s and %s with %s as peer"
		    , ocb, pcb, pb);
	    }
	    else
	    {
		/* We will only request an IPsec SA if policy isn't empty
		 * (ignoring Main Mode items).
		 * This is a fudge, but not yet important.
		 * If we are to proceed asynchronously, whackfd will be NULL_FD.
		 */
		passert(c->gw_info != NULL);
		c->gw_info->last_tried_time = now();
		ipsecdoi_initiate(whackfd
		    , c
		    , (c->policy>>POLICY_IPSEC_SHIFT) != 0
		    , c->policy
		    , 1);
		whackfd = NULL_FD;	/* protect from close */
	    }
	}
    }
    if (whackfd != NULL_FD)
	close(whackfd);
}
void
terminate_connection(const char *nm)
{
    /* Loop because more than one may match (master and instances)
     * But at least one is required (enforced by con_by_name).
     */
    struct connection *c, *n;

    for (c = con_by_name(nm, TRUE); c != NULL; c = n)
    {
	n = c->next;	/* grab this before c might disappear */
	if (streq(c->name, nm))
	{
	    SET_CUR_CONNECTION(c);
	    log("terminating SAs using connection");
	    delete_states_by_connection(c);
	    UNSET_CUR_CONNECTION();
	}
    }
}

/* check nexthop safety
 * Our nexthop must not be within a routed client subnet, and vice versa.
 * Note: we don't think this is true.  We think that KLIPS will
 * not process a packet output by an eroute.
 */
#ifdef NEVER
bool
check_nexthop(const struct connection *c)
{
    struct connection *d;

    if (addrinsubnet(&c->this.host_nexthop, &c->that.client))
    {
	loglog(RC_LOG_SERIOUS, "cannot perform routing for connection \"%s\""
	    " because nexthop is within peer's client network",
	    c->name);
	return FALSE;
    }

    for (d = connections; d != NULL; d = d->next)
    {
	if (d->routed)
	{
	    if (addrinsubnet(&c->this.host_nexthop, &d->that.client))
	    {
		loglog(RC_LOG_SERIOUS, "cannot do routing for connection \"%s\"
		    " because nexthop is contained in"
		    " existing routing for connection \"%s\"",
		    c->name, d->name);
		return FALSE;
	    }
	    if (addrinsubnet(&d->this.host_nexthop, &c->that.client))
	    {
		loglog(RC_LOG_SERIOUS, "cannot do routing for connection \"%s\"
		    " because it contains nexthop of"
		    " existing routing for connection \"%s\"",
		    c->name, d->name);
		return FALSE;
	    }
	}
    }
    return TRUE;
}
#endif /* NEVER */

/* an ISAKMP SA has been established.
 * Note the serial number, and unorient any connections with
 * the same peer ID but different peer IP address.
 */
bool uniqueIDs = FALSE;	/* --uniqueids? */

void
ISAKMP_SA_established(struct connection *c, so_serial_t serial)
{
    c->newest_isakmp_sa = serial;

    if (uniqueIDs)
    {
	/* for all connections: if the same Phase 1 peer ID is used
	 * for a different IP address, unorient that connection.
	 */
	struct connection *d;

	for (d = connections; d != NULL; )
	{
	    struct connection *next = d->next;	/* might move underneath us */

	    if (same_id(&c->that.id, &d->that.id)
	    && !sameaddr(&c->that.host_addr, &d->that.host_addr)
	    && !HasWildcardIP(*d))
	    {
		unorient_connection(d);
	    }
	    d = next;
	}
	
    }
}

/* Find a connection that owns the route to a connection's peer's client.
 * Preference is given to one that has an eroute too.
 * If eroute is true, our client must match too (as for eroutes).
 */
struct connection *
route_owner(struct connection *c, bool eroute)
{
    struct connection
	*d,
	*best = NULL;

    passert(oriented(*c));

    if (c->routed && c->eroute_owner != SOS_NOBODY)
	return c;

    for (d = connections; d != NULL; d = d->next)
    {
	if (d->routed)
	{
	    passert(oriented(*d));
	    if (samesubnet(&c->that.client, &d->that.client)
	    && (!eroute || samesubnet(&c->this.client, &d->this.client)))
	    {
		if (d->eroute_owner != SOS_NOBODY)
		    return d;	/* definite winner */

		best = d;	/* winner if no better comes along */
	    }
	}
    }
    return best;
}

struct connection *
find_host_connection(const ip_address *me, u_int16_t my_port
, const ip_address *him, u_int16_t his_port)
{
    struct connection *c = find_host_pair_connections(me, my_port, him, his_port);

    if (c != NULL && !oriented(*c))
	(void)orient(c, FALSE);
    return c;
}

/* given an up-until-now satisfactory connection, find the best
 * connection now that we've got the Id Payload from the peer.
 *
 * We can only accept connections with compatible authentication.
 * The authentication method has been committed to.
 * The authentication materials are a slightly different matter.
 * We are committed to all authentication materials EXCEPT
 * in the case of RSA Sig authentication when we are the Responder.
 * In that case, the Initiator's public key is still open.
 */
struct connection *
refine_host_connection(const struct state *st, const struct id *peer_id
, bool initiator)
{
    struct connection *c = st->st_connection;
    u_int16_t auth = st->st_oakley.auth;
    struct connection *d;
    lset_t auth_policy;
    const chunk_t *psk;
    const struct RSA_public_key *his_RSA_pub;
    const struct RSA_private_key *my_RSA_pri;
    bool rw;

    if (same_id(&c->that.id, peer_id))
	return c;

    switch (auth)
    {
    case OAKLEY_PRESHARED_KEY:
	auth_policy = POLICY_PSK;
	psk = get_preshared_secret(c);
	if (psk == NULL)
	    return NULL;	/* cannot determine PSK! */
	break;
    case OAKLEY_RSA_SIG:
	auth_policy = POLICY_RSASIG;
	my_RSA_pri = get_RSA_private_key(c);
	if (my_RSA_pri == NULL)
	    return NULL;	/* cannot determine my RSA private key! */
	if (initiator)
	{
	    his_RSA_pub = get_his_RSA_public_key(c);
	    if (his_RSA_pub == NULL)
		return NULL;	/* cannot determine his RSA public key! */
	}
	break;
    default:
	passert(FALSE);
    }
    /* The current connection won't do: search for one that will.
     * First search for one with the same pair of hosts.
     * If that fails, search for one with a Road Warrior peer.
     */
    d = c->host_pair->connections;
    for (rw = FALSE; ; rw = TRUE)
    {
	struct connection *best_found = NULL;

	for (; d != NULL; d = d->hp_next)
	{
	    bool exact;

	    if (!oriented(*d))
		(void)orient(d, FALSE);

	    if (!same_id(&c->this.id, &d->this.id))
		continue;

	    exact = same_id(peer_id, &d->that.id);

	    if (exact || (rw && d->that.id.kind == ID_NONE))
	    {
		/* check whether authentication would work the same */

		if (!((d->policy & auth_policy) != 0
		|| (d->policy & POLICY_ISAKMP_MASK) == 0))
		    continue;	/* our auth isn't OK for this connection */

		switch (auth)
		{
		case OAKLEY_PRESHARED_KEY:
		    {
			const chunk_t *dpsk = get_preshared_secret(d);

			if (dpsk == NULL)
			    continue;	/* no secret */

			if (psk != dpsk)
			    if (psk->len != dpsk->len
			    || memcmp(psk->ptr, dpsk->ptr, psk->len) != 0)
				continue;	/* different secret */
		    }
		    break;
		case OAKLEY_RSA_SIG:
		    if (my_RSA_pri != get_RSA_private_key(d))
			continue;	/* different private key */

		    if (initiator)
		    {
			/* his public key has been used for authentication
			 * so we must be sure that any connection we
			 * select will use the same public key.
			 */
			const struct RSA_public_key *hnk = exact
			    ? get_his_RSA_public_key(d) : get_RSA_public_key(peer_id);

			passert(exact || d->gw_info == NULL);

			if (!same_RSA_public_key(his_RSA_pub, hnk))
			    continue;	/* different public key */
		    }
		    break;
		default:
		    passert(FALSE);
		}

		if (exact)
		    return d;	/* passed all the tests */

		best_found = d;	/* passed tests, but peer_id was wildcarded */
	    }
	}
	if (rw)
	    return best_found;	/* been around twice already */

	/* for second time around: figure out RW host pair's connections */
	d = find_host_pair_connections(&c->this.host_addr, c->this.host_port
	    , (ip_address *)NULL, c->that.host_port);
    }
}

/* find_client_connection: given a connection suitable for ISAKMP
 * (i.e. the hosts match), find a one suitable for IPSEC
 * (i.e. with matching clients).
 *
 * If we don't find an exact match (not even our current connection),
 * we try for one that still needs instantiation.  Try Road Warrior
 * abstract connections and the Opportunistic abstract connections.
 * This requires inverse instantiation: abstraction.
 *
 * After failing to find an exact match, we abstract the peer
 * to be NO_IP (the wildcard value).  This enables matches with
 * Road Warrior and Opportunistic abstract connections.
 *
 * After failing that search, we also abstract the Phase 1 peer ID
 * if possible.  If the peer's ID was the peer's IP address, we make
 * it NO_ID; instantiation will make it the peer's IP address again.
 *
 * If searching for a Road Warrior abstract connection fails,
 * and conditions are suitable, we search for an Opportunistic
 * abstract connection.
 *
 * Note: in the end, both Phase 1 IDs must be preserved, after any
 * instantiation.  They are the IDs that have been authenticated.
 */

/* fc_try: a helper function for find_client_connection */
static struct connection *
fc_try(bool oppo
, const struct connection *c
, struct host_pair *hp
, const struct id *peer_id
, const ip_subnet *our_net
, const ip_subnet *peer_net)
{
    struct connection *d;
    struct connection *unrouted = NULL;
    const bool peer_net_is_host = subnetishost(peer_net)
	&& addrinsubnet(&c->that.host_addr, peer_net);

    for (d = hp->connections; d != NULL; d = d->hp_next)
    {
	if (!oriented(*d))
	    (void)orient(d, FALSE);

	if (!same_peer_ids(c, d, peer_id))
	    continue;

	if (oppo)
	{
	    /* Opportunistic case:
	     * our_client must be inside the abstract client
	     * and we must have a Wildcard Client.
	     */
	    if (!addrinsubnet(&d->this.host_addr, our_net)
	    || !HasWildcardClient(c))
		continue;
	}
	else
	{
	    /* non-Opportunistic case:
	     * our_client must match.
	     *
	     * So must peer_client, but the testing is complicated
	     * by the fact that the peer might be a wildcard
	     * and if so, the default value of that.client
	     * won't match the default peer_net.  The appropriate test:
	     * 
	     * If d has a peer client, it must match peer_net.
	     * If d has no peer client, peer_net must just have peer itself.
	     */
	    if (!samesubnet(&d->this.client, our_net))
		continue;

	    if (d->that.has_client)
	    {
		if (!samesubnet(&d->that.client, peer_net))
		    continue;
	    }
	    else
	    {
		if (!peer_net_is_host)
		    continue;
	    }
	}

	/* we've run the gauntlet -- success */
	if (d->routed)
	    return d;

	if (unrouted == NULL)
	    unrouted = d;
    }
    return unrouted;
}

struct connection *
find_client_connection(struct connection *c
, const ip_subnet *our_net, const ip_subnet *peer_net)
{
    struct connection *d;

    /* give priority to current connection
     * but even greater priority to a routed concrete connection
     */
    {
	struct connection *unrouted = NULL;

	if (samesubnet(&c->this.client, our_net)
	&& samesubnet(&c->that.client, peer_net))
	{
	    passert(oriented(*c));
	    if (c->routed)
		return c;

	    unrouted = c;
	}

	/* exact match? */
	d = fc_try(FALSE, c, c->host_pair, NULL, our_net, peer_net);
	if (d == NULL)
	    d = unrouted;
    }

    if (d == NULL)
    {
	/* look for an abstract connection to match */
	struct id abstract_peer_id;
	struct host_pair *hp = find_host_pair(&c->this.host_addr
	    , c->this.host_port, NULL, c->that.host_port);

	abstract_peer_id.kind = ID_NONE;

	/* RW match with actual peer_id? */
	d = fc_try(FALSE, c, hp, NULL, our_net, peer_net);

	if (d == NULL && (c->that.id.kind == ID_IPV4_ADDR || c->that.id.kind == ID_IPV6_ADDR)
	&& sameaddr(&c->that.id.ip_addr, &c->that.host_addr))
	{
	    /* RW match with abstract peer_id?
	     * Note that later instantiation will result in the same peer_id.
	     */
	    d = fc_try(FALSE, c, hp, &abstract_peer_id, our_net, peer_net);
	}

	if (d == NULL && subnetishost(our_net)
	&& subnetishost(peer_net))
	{
	    /* Opportunistic match?
	     * Always use abstract peer_id.
	     * Note that later instantiation will result in the same peer_id.
	     */
	    d = fc_try(TRUE, c, hp, &abstract_peer_id, our_net, peer_net);
	}
    }

    return d;
}

void
show_connections_status(void)
{
    struct connection *c;

    for (c = connections; c != NULL; c = c->next)
    {
	const char *routed = c->routed? "; routed" : "";
	const char *ifn = oriented(*c)? c->interface->rname : "";
	const char *instance = c->rw_state == rwcs_instance? " instance" : "";

	/* show topology */
	{
	    char lhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];
	    size_t ll = format_end(lhs, sizeof(lhs), &c->this, &c->that, TRUE);
	    char rhs[SUBNETTOT_BUF + ADDRTOT_BUF + IDTOA_BUF + ADDRTOT_BUF];
	    size_t rl = format_end(rhs, sizeof(rhs), &c->that, &c->this, FALSE);

	    if (ll + rl < 70)
	    {
		/* display on one line */
		whack_log(RC_COMMENT, "\"%s\"%s: %s...%s"
		    , c->name, instance, lhs, rhs);
	    }
	    else
	    {
		/* split over two lines */
		whack_log(RC_COMMENT, "\"%s\"%s: %s..."
		    , c->name, instance, lhs);
		whack_log(RC_COMMENT, "\"%s\"%s: ...%s"
		    , c->name, instance, rhs);
	    }
	}

	whack_log(RC_COMMENT
	    , "\"%s\"%s:   ike_life: %lus; ipsec_life: %lus;"
	    " rekey_margin: %lus; rekey_fuzz: %lu%%; keyingtries: %lu"
	    , c->name
	    , instance
	    , (unsigned long) c->sa_ike_life_seconds
	    , (unsigned long) c->sa_ipsec_life_seconds
	    , (unsigned long) c->sa_rekey_margin
	    , (unsigned long) c->sa_rekey_fuzz
	    , (unsigned long) c->sa_keying_tries);

	whack_log(RC_COMMENT
	    , "\"%s\"%s:   policy: %s; interface: %s%s"
	    , c->name
	    , instance
	    , bitnamesof(sa_policy_bit_names, c->policy)
	    , ifn
	    , routed);

	whack_log(RC_COMMENT
	    , "\"%s\"%s:   newest ISAKMP SA: #%ld; newest IPsec SA: #%ld; eroute owner: #%lu"
	    , c->name
	    , instance
	    , c->newest_isakmp_sa
	    , c->newest_ipsec_sa
	    , c->eroute_owner);
    }
}

/* Is c->that.client 0.0.0.0/32 (or IPV6 equivalent)?
 * This, along with c->that.host being 0.0.0.0 signifies Opportunism.
 */
bool
HasWildcardClient(const struct connection *c)
{
    if (c->that.has_client && subnetishost(&c->that.client))
    {
	ip_address addr;

	networkof(&c->that.client, &addr);
	if (isanyaddr(&addr))
	    return TRUE;
    }
    return FALSE;
}
