/*
 *  linux/fs/supermount/dir.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 *  $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 */

#include <asm/segment.h>
#include <asm/current.h>

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/stat.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/poll.h>

static int supermount_dir_open (struct inode *, struct file *);
static loff_t supermount_dir_lseek (struct file *, loff_t, int);
static ssize_t supermount_dir_read (struct file *, char *, size_t, loff_t *);
static int supermount_readdir (struct file *, void *, filldir_t);
static int supermount_dir_ioctl (struct inode *, struct file *, unsigned int, 
							 unsigned long);
static int supermount_dir_flush (struct file *);
static int supermount_dir_fsync (struct file *, struct dentry *);
static int supermount_dir_release (struct inode *, struct file *);
static int supermount_dir_fasync (int, struct file *, int);

static struct file_operations supermount_dir_operations = {
	supermount_dir_lseek,	/* lseek */
	supermount_dir_read,	/* read */
	NULL,				/* write - bad */
	supermount_readdir,	/* readdir */
	NULL,				/* poll */
	supermount_dir_ioctl,	/* ioctl */
	NULL,				/* mmap */
	supermount_dir_open,	/* open */
	supermount_dir_flush,	/* flush */
	supermount_dir_release,	/* release */
	supermount_dir_fsync,	/* fsync */
	supermount_dir_fasync,	/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

/*
 * directories can handle most operations...  supermount/namei.c just
 * passes them through to the underlying subfs, except for lookup().
 */
struct inode_operations supermount_dir_iops = {
	&supermount_dir_operations,	/* default directory file-ops */
	supermount_create,		/* create */
	supermount_lookup,		/* lookup */
	NULL/*supermount_link*/,		/* link */
	supermount_unlink,		/* unlink */
	supermount_symlink,		/* symlink */
	supermount_mkdir,		/* mkdir */
	supermount_rmdir,		/* rmdir */
	supermount_mknod,		/* mknod */
	supermount_rename,		/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	NULL,				/* readpage */
	NULL,				/* writepage */
	NULL,				/* bmap */
	NULL,				/* truncate */
	supermount_permission,		/* permission */
	NULL,				/* swap */
	NULL,				/* updatepage */
	NULL				/* revalidate */
};

/* 
 * When we open() a file, just shadow the file with
 * a file open to the subinode, and let that handle all the work.  
 */

int supermount_dir_open (struct inode * inode, struct file * file)
{
	struct inode * subi;
	struct dentry *subdent;
	struct file * subfile;
	
	int rc;

	supermount_debug("pwd: %08lx, %08lx, %d, %s\n",
		(unsigned long)current->fs->pwd,
		(unsigned long)current->fs->pwd->d_parent,
		list_empty(&current->fs->pwd->d_hash),
		current->fs->pwd->d_name.name);
	supermount_debug("open inode %ld, name: %s\n", inode->i_ino, file->f_dentry->d_name.name);
	supermount_debug("File dentry count: %d\n",
		file->f_dentry->d_count);

	if (S_ISDIR(inode->i_mode) && (file->f_mode & FMODE_WRITE))
		return -EPERM;

	subfile = get_empty_filp();
	if (!subfile)
		return -ENFILE;

	subi = subiget(inode);
	if (!subi) {
		put_filp(subfile);
		return -ENOENT;
	}
	supermount_iopen(inode->i_sb);
	supermount_dump_inode(inode);

	if (file->f_mode & FMODE_WRITE)
	{
		rc = supermount_get_write_access (inode->i_sb);
		if (rc)
		{
			put_filp(subfile);
			iput(subi);
			supermount_iclose(inode->i_sb);
			return rc;
		}
	}

	subdent = dget(inode->u.supermount_i.i_dentry);
	if (!subdent) {
		put_filp(subfile);
		iput(subi);
		supermount_iclose(inode->i_sb);

		return -ENOENT;
	}

	file->private_data = (void *) subfile;
	
	subfile->f_dentry = subdent;
	subfile->f_op = subi->i_op ? subi->i_op->default_file_ops : NULL;
	subfile->f_pos = 0;
	subfile->f_reada = 0;
	subfile->f_flags = file->f_flags;
	subfile->f_mode = file->f_mode;

	iput(subi); /* not needed anymore */
	supermount_dump_normal_inode(subi);
	
	if (subfile->f_op && subfile->f_op->open) {
		rc = subfile->f_op->open(subi, subfile);
		if (rc) {
			dput(subdent);
			put_filp(subfile);
			return rc;
		}
	}
	return 0;
}

/*
 * Pass on an lseek() to the underlying subfs
 */

static loff_t supermount_dir_lseek (struct file *file, loff_t offset, int origin)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode = subfile->f_dentry->d_inode;
	loff_t retval;

	supermount_debug ("dir_lseek inode %ld\n", file->f_dentry->d_inode->i_ino);

	if (!subinode) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_lseek", 
				  "No hidden inode on open directory");
	}
	
	retval = do_lseek(subfile, offset, origin);
	file->f_pos = subfile->f_pos;
	file->f_reada = subfile->f_reada;
	file->f_version = subfile->f_version;

	return retval;
}

static ssize_t supermount_dir_read (struct file *file, char *buf, size_t count,
								loff_t *ppos)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode = subfile->f_dentry->d_inode;

	supermount_debug ("read inode %ld, subinode: %ld/%ld\n",
		file->f_dentry->d_inode->i_ino, subinode->i_ino,
		subfile->f_dentry->d_inode->i_ino);

	if (!subinode) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_dir_read", 
				  "No hidden inode on open inode");
	}

	if (subfile->f_op && subfile->f_op->read)
		return subfile->f_op->read(subfile, buf, count, ppos);

	return -EINVAL;
}

struct inode *get_possible_super_inode(struct super_block *sb, ino_t ino);

/*
 * All magic in this function is because of obsolete inodes still present in
 * filesystem
 */
int supermount_readdir (struct file *file, void * buf,
							   filldir_t fill_fn)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode = subfile->f_dentry->d_inode;
	int supermount_filldir_callback(void * __buf, const char * name, int namlen, off_t offset, ino_t ino)
	{
		struct qstr qname = {name, namlen, full_name_hash(name, namlen)};
		ino_t tmp_ino = find_inode_number(file->f_dentry, &qname);
		struct inode *tmp;
/*		supermount_debug("%d, %*s\n", namlen, namlen, name);*/
		if (tmp_ino)
		{
			supermount_debug("found superinode: %ld/%ld/%ld\n",
				file->f_dentry->d_inode->i_ino, ino, tmp_ino);
			ino = tmp_ino;
		}
		else
		{
			tmp = get_possible_super_inode(file->f_dentry->d_sb, ino);
			supermount_debug("new superinode: %ld/%ld/%ld\n",
				file->f_dentry->d_inode->i_ino, ino, tmp->i_ino);
			ino = tmp->i_ino;
			iput(tmp);
		}
		return fill_fn(__buf, name, namlen, offset, ino, DT_DIR);
	}

	if (!S_ISDIR(file->f_dentry->d_inode->i_mode))
		return -EINVAL;
	
	supermount_debug ("readdir inode %ld, subinode: %ld\n",
		file->f_dentry->d_inode->i_ino, subinode->i_ino);

	if (!subinode) {
		supermount_panic (file->f_dentry->d_inode->i_sb, "supermount_readdir", 
				  "No hidden inode on open directory");
	}
	
	if (subfile->f_op && subfile->f_op->readdir)
		return subfile->f_op->readdir(subfile, buf, supermount_filldir_callback);

	return -EINVAL;
}

static int supermount_dir_ioctl (struct inode *inode, struct file *file,
				 unsigned int cmd, unsigned long arg)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode = subfile->f_dentry->d_inode;

	supermount_debug ("dir_ioctl inode %ld\n", inode->i_ino);

	if (subfile->f_op && subfile->f_op->ioctl)
		return subfile->f_op->ioctl(subinode, subfile, cmd, arg);

	return -ENOTTY;
}

static int supermount_dir_flush (struct file *file)
{
	struct file *subfile = (struct file *) file->private_data;

	supermount_dump_inode(file->f_dentry->d_inode);

	if (subfile->f_op && subfile->f_op->flush)
		return subfile->f_op->flush(subfile);

	return 0;
}

static int supermount_dir_fsync (struct file *file, struct dentry *dentry)
{
	struct file *subfile = (struct file *) file->private_data;

	supermount_dump_inode(file->f_dentry->d_inode);

	if (subfile->f_op && subfile->f_op->fsync)
		return subfile->f_op->fsync(subfile, subfile->f_dentry);

	return -EINVAL;
}

static int supermount_dir_fasync (int fd, struct file *file, int on)
{
	struct file *subfile = (struct file *) file->private_data;

	supermount_dump_inode(file->f_dentry->d_inode);

	if (subfile->f_op && subfile->f_op->fasync)
		return subfile->f_op->fasync(fd, subfile, on);

	return -EINVAL;
}

static int supermount_dir_release (struct inode *inode, struct file *file)
{
	struct file *subfile = (struct file *) file->private_data;
	struct inode *subinode = subfile->f_dentry->d_inode;

	supermount_debug ("release inode %ld\n", inode->i_ino);

	if (!subinode) {
		supermount_panic (inode->i_sb, "supermount_dir_ioctl", 
				  "No hidden inode on open directory");
	}

	supermount_debug("Root count: %d\n", inode->i_sb->s_root->d_count);
	supermount_debug("Subroot count: %d\n",
		inode->i_sb->u.supermount_sb.s_subfs->s_root->d_count);
	supermount_dump_normal_inode(inode->i_sb->u.supermount_sb.s_subfs->s_root->d_inode);
	supermount_debug("Subfile dentry: %08lx\n", (unsigned long)subfile->f_dentry);

	filp_close (subfile, current->files);
	if (file->f_mode & FMODE_WRITE)
	{
		mark_subfs_dirty(inode->i_sb);
		supermount_put_write_access (inode->i_sb);
	}
	supermount_iclose(inode->i_sb);

	supermount_debug("Subroot count: %d\n",
		inode->i_sb->u.supermount_sb.s_subfs->s_root->d_count);
	supermount_dump_normal_inode(inode->i_sb->u.supermount_sb.s_subfs->s_root->d_inode);
	supermount_debug("Subfile dentry: %08lx\n", (unsigned long)subfile->f_dentry);

	supermount_dump_inode(inode);

	supermount_debug("File dentry count: %d, list: %d/%d\n",
		file->f_dentry->d_count,
		list_empty(&file->f_dentry->d_lru),
		list_empty(&file->f_dentry->d_hash)
		);

	return 0;
}

static void dir_dentry_iput(struct dentry *dentry, struct inode *inode)
{
	struct dentry *subd;

	supermount_debug("dentry %s(%08x/%08x), iput: %ld\n", dentry->d_name.name,
		(unsigned int)dentry, (unsigned int)dentry->d_parent, inode->i_ino);
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_iput)
	{
		supermount_debug("lowlevel\n");
		subd->d_op->d_iput(subd, inode);
		return;
	}
	iput(inode);
}

static void dir_dentry_delete(struct dentry *dentry)
{
	struct dentry *subd;

	supermount_debug("delete dir dentry %s(%08x/%08x)\n", dentry->d_name.name,
		(unsigned int)dentry, (unsigned int)dentry->d_parent);
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_delete)
	{
		supermount_debug("lowlevel\n");
		subd->d_op->d_delete(subd);
	}
}

static void dir_dentry_release(struct dentry *dentry)
{
	struct dentry *subd;

	supermount_debug("release dir dentry %s(%08x/%08x)\n", dentry->d_name.name,
		(unsigned int)dentry, (unsigned int)dentry->d_parent);
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_release)
	{
		supermount_debug("lowlevel\n");
		subd->d_op->d_release(subd);
	}
}

static int dir_dentry_revalidate(struct dentry *dentry, int flags)
{
	struct dentry *subd;

	supermount_debug("revalidate dir dentry %s\n", dentry->d_name.name);
	if (!supermount_revalidate_inode(dentry->d_sb, dentry->d_inode))
	{
		d_drop(dentry);
		return 0;
	}
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_revalidate)
	{
		supermount_debug("lowlevel\n");
		return subd->d_op->d_revalidate(subd, flags);
	}
	return 1;
}

static int dir_dentry_compare(struct dentry *dentry, struct qstr *name1,
	struct qstr *name2)
{
	struct dentry *subd;

	supermount_debug("compare dir dentry %s, %s, %s\n", dentry->d_name.name,
		name1->name, name2->name);
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_compare)
	{
		supermount_debug("lowlevel\n");
		return subd->d_op->d_compare(subd, name1, name2);
	}

	if (name1->len != name2->len)
		return 1;
	return memcmp(name1->name, name2->name, name1->len);
}

static int dir_dentry_hash(struct dentry *dentry, struct qstr *name)
{
	struct dentry *subd;

	supermount_debug("hash dir dentry %s, %s\n", dentry->d_name.name,
		name->name);
	if (dentry->d_inode &&
			(subd = dentry->d_inode->u.supermount_i.i_dentry) != 0 &&
			subd->d_op && subd->d_op->d_hash)
	{
		supermount_debug("lowlevel\n");
		return subd->d_op->d_hash(subd, name);
	}

	return 0;
}

struct dentry_operations supermount_dir_dops = {
	dir_dentry_revalidate,
	dir_dentry_hash,
	dir_dentry_compare,
	dir_dentry_delete,
	dir_dentry_release,
	dir_dentry_iput
};
