/*
 *  linux/fs/supermount/file.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 * $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 */

#include <linux/fs.h>
#include <linux/supermount_fs.h>

static int file_revalidate_dentry(struct dentry *dentry, int flags)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file revalidate dentry %s\n", dentry->d_name.name);

	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		if (!supermount_revalidate_inode(inode->i_sb, inode))
		{
			d_drop(dentry);
			return 0;
		}
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_revalidate)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_revalidate(dentry, flags);
		}
	}
		
	return 1;
}

static int file_hash_dentry(struct dentry *dentry, struct qstr *name)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file hash dentry %s\n", dentry->d_name.name);
	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_hash)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_hash(dentry, name);
		}
	}

	return 0;
}

static int file_compare_dentry(struct dentry *dentry, struct qstr *name1,
	struct qstr *name2)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file compare dentry %s\n", dentry->d_name.name);
	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_compare)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_compare(dentry, name1, name2);
		}
	}

	if (name1->len != name2->len)
		return 1;
	return memcmp(name1->name, name2->name, name1->len);
}

static void file_delete_dentry(struct dentry *dentry)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file delete dentry %s\n", dentry->d_name.name);
	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_delete)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_delete(dentry);
		}
	}
}

static void file_release_dentry(struct dentry *dentry)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file release dentry %s\n", dentry->d_name.name);
	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_release)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_release(dentry);
		}
	}
}

static void file_iput_dentry(struct dentry *dentry, struct inode *subi)
{
	struct inode *inode;
	struct dentry *subd;

	supermount_debug("file iput dentry %s\n", dentry->d_name.name);
	if (dentry->d_inode && (inode = dentry->d_inode->i_shadow) != 0)
	{
		subd = inode->u.supermount_i.i_dentry;
		if (subd && subd->d_op && subd->d_op->d_iput)
		{
			supermount_debug("lowlevel\n");
			return subd->d_op->d_iput(dentry, subi);
		}
	}

	iput(subi);
}

struct dentry_operations supermount_file_dops = {
	file_revalidate_dentry,
	file_hash_dentry,
	file_compare_dentry,
	file_delete_dentry,
	file_release_dentry,
	file_iput_dentry
};
