/*
 *  linux/fs/supermount/inode.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/inode.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *      and           1993  Stephen Tweedie
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 * $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 */
#include <asm/segment.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>

spinlock_t hidden_lock = SPIN_LOCK_UNLOCKED;

/*
 * We've got a superinode here: do an iget on the appropriate subfs inode 
 * 
 * Revalidate the media first if necessary, and return NULL if we can't
 * get a subinode (media may no longer be loaded).
 *
 * If superinode is obsolete, return inode of subfs root as subinode. Don't
 * create hidden subinode. Besides, as we are building a tree of subfs
 * dentries in superinode->u.supermount_i.i_dentry, we need .i_dentry of
 * this superinode to be used for creating of subfile in direcory operations
 * and as parent for creating .i_dentry of subdirectories of this directory
 * in supermount_lookup. So if we don't have
 * superinode->u.supermount_i.i_dentry we create it as root dentry with
 * inode of subfs's root.
 *
 * If superinode is obsolete d_drop it's dentry, so futher attempt to use
 * this entry will be forced to do full lookup and therefore get new inode.
 */

struct inode * subiget(struct inode * inode) 
{
	struct inode * tmp;
	supermount_debug("subiget inode %ld\n", inode->i_ino);

/*	if (!S_ISDIR(inode->i_mode))
		return 0;*/
	if (supermount_media_check(inode->i_sb)) {
		supermount_debug("media is invalid\n");
		return 0;
	}

	supermount_dump_inode(inode);

	if (!subfs_is_active(inode->i_sb)) {
		supermount_debug ("going online.\n");
		inode->i_sb->u.supermount_sb.s_state = 
			SUPERMOUNT_ONLINE;
	}

	if (inode_is_obsolete(inode)) {
		/* Drop dentry */
		if (!list_empty(&inode->i_dentry))
			d_drop(list_entry(inode->i_dentry.next, struct dentry, d_alias));

		/* For obsolete directory with new mounted media ---
                   return the new root subinode. */
		if (inode->u.supermount_i.i_hidden) 
			supermount_panic (inode->i_sb, "subiget", 
					  "Hidden inode on obsolete inode %ld",
					  inode->i_ino);

		tmp = inode->i_sb->u.supermount_sb.s_subfs->s_root->d_inode;
		tmp->i_count++;
		inode->u.supermount_i.i_subino = tmp->i_ino;

		if (!inode->u.supermount_i.i_dentry) 
		{
			tmp->i_count++; /* in use by dentry now */
			inode->u.supermount_i.i_dentry = d_alloc_root(tmp, NULL);
			supermount_debug("new dentry: %08x\n",
				(unsigned int) inode->u.supermount_i.i_dentry);
		}

		supermount_debug("mapping obsolete inode %ld to "
				 "subroot inode %ld\n",
				 inode->i_ino, tmp->i_ino);
		return tmp;
	}

	if (inode->u.supermount_i.i_hidden) {
		inode->u.supermount_i.i_hidden->i_count++;
		supermount_debug("inode %ld/%ld, found hidden, "
				 "count now %d/%d\n",
				 inode->i_ino,
				 inode->u.supermount_i.i_hidden->i_ino,
				 inode->i_count,
				 inode->u.supermount_i.i_hidden->i_count);
		return inode->u.supermount_i.i_hidden;
	}

	supermount_read_hidden_inode(inode);
	if (!inode->u.supermount_i.i_hidden) {
		supermount_debug("inode %ld, no subinode, count=%d\n",
				 inode->i_ino, inode->i_count);
		return 0;
	}
	
	supermount_debug("inode %ld/%ld, found new, "
			 "count now %d/%d\n",
			 inode->i_ino,
			 inode->u.supermount_i.i_hidden->i_ino,
			 inode->i_count,
			 inode->u.supermount_i.i_hidden->i_count);
	return inode->u.supermount_i.i_hidden;
}

/* Supermount open/close inode.  These functions maintain the
   superblock reference counts of active inodes; the subfs is
   suspended when that count reaches zero.

   Files are opened on any significant reference, and are not closed
   until they become fully dereferenced (during the last iput).
   Directories, on the other hand, are always closed unless they are
   active and opened; a directory referenced as CWD is not open.

   The difference is in the way the application gets given inodes.
   For dirs, it will be given the superinode to deal with; for other
   file types, we return the subinode. */


void supermount_iopen(struct super_block * sb)
{
	switch (sb->u.supermount_sb.s_state) {
	case SUPERMOUNT_UNMOUNTED:
		supermount_panic (sb, "supermount_iopen", 
				  "opening inode on unmounted subfs");
	case SUPERMOUNT_SUSPENDED:
		supermount_debug("locking door.\n");
		supermount_lock_door(sb, 1);
		break;
	default: 
	}
	
	if (!subfs_is_active(sb)) {
		supermount_debug ("going online.\n");
		sb->u.supermount_sb.s_state = 
			SUPERMOUNT_ONLINE;
	}
	sb->u.supermount_sb.s_opencount++;
	supermount_debug("Opencount: %d\n", sb->u.supermount_sb.s_opencount);
}

void supermount_iclose(struct super_block * sb)
{
	sb->u.supermount_sb.s_opencount--;
	supermount_debug("Opencount: %d\n", sb->u.supermount_sb.s_opencount);
	if (!subfs_is_active(sb))
		supermount_go_inactive(sb);
}

void supermount_read_inode (struct inode * inode)
{
	supermount_dump_inode(inode);
	
	inode->i_shadow_op = &supermount_shadow_iops;
	inode->u.supermount_i.i_sb_version = 
		inode->i_sb->u.supermount_sb.s_version;
	
	switch (inode->i_ino) {
	case SUPERMOUNT_ROOT_INO:
		if (inode->i_sb->u.supermount_sb.s_state !=
		    SUPERMOUNT_UNMOUNTED) {
			supermount_panic (inode->i_sb, "supermount_read_inode",
					  "Help - trying to read root while "
					  "already mounted!");
		}
		/* Fall through */
	case SUPERMOUNT_HIDDEN_INO:
		inode->i_mode = inode->i_sb->u.supermount_sb.s_default_mode;
		inode->i_uid = 0;
		inode->i_gid = 0;
		inode->i_nlink = 1;
		inode->i_size = 0;
		inode->i_atime = CURRENT_TIME;
		inode->i_ctime = CURRENT_TIME;
		inode->i_mtime = CURRENT_TIME;
		inode->i_blksize = inode->i_sb->s_blocksize;
		inode->i_blocks = 0;
		inode->i_version = ++global_event;
		inode->i_op = (inode->i_ino == SUPERMOUNT_ROOT_INO)
			? &supermount_dir_iops : NULL;
		return;
	default:
		;
	}
}

void supermount_write_inode (struct inode * inode)
{
	struct super_block *tmp;
	tmp = inode->i_sb->u.supermount_sb.s_subfs;
	
	supermount_dump_inode(inode);

	if (inode->u.supermount_i.i_hidden &&
		    tmp && tmp->s_op && tmp->s_op->write_inode)
	{
		tmp->s_op->write_inode(inode->u.supermount_i.i_hidden);
		inode->i_mode = inode->u.supermount_i.i_hidden->i_mode;
		inode->i_uid = inode->u.supermount_i.i_hidden->i_uid;
		inode->i_gid = inode->u.supermount_i.i_hidden->i_gid;
		inode->i_size = inode->u.supermount_i.i_hidden->i_size;
		inode->i_atime = inode->u.supermount_i.i_hidden->i_atime;
		inode->i_mtime = inode->u.supermount_i.i_hidden->i_mtime;
		inode->i_ctime = inode->u.supermount_i.i_hidden->i_ctime;
	}

	mark_subfs_dirty(inode->i_sb);
}

/*
 * One special case here, when superinode->i_count == 2, superinode linked to
 * subinode and hidden dentry present. This means that then only holder of
 * superinode's reference is subinode link. In order to break link we need to
 * iput subinode, which is in use by hidden dentry. So we dput hidden dentry.
 */

void supermount_put_inode (struct inode * inode)
{
	supermount_dump_inode(inode);

	if (inode->i_count == 2 && inode->u.supermount_i.i_dentry &&
		(inode->u.supermount_i.i_dentry->d_inode == inode ||
			inode->u.supermount_i.i_dentry->d_inode == 
				inode->u.supermount_i.i_hidden))
	{
		struct dentry *tmp = inode->u.supermount_i.i_dentry;
		inode->u.supermount_i.i_dentry = 0;
		if (tmp)
		{
			supermount_debug("hidden_dentry %s, count %d\n", tmp->d_name.name, tmp->d_count);
			if (tmp->d_inode)
				supermount_dump_normal_inode(tmp->d_inode);
		}
		dput(tmp);
	}	

	if (inode->i_count != 1)
		return;

	supermount_debug("hidden dentry: %08lx\n", (unsigned long)inode->u.supermount_i.i_dentry);
	if (inode->u.supermount_i.i_dentry)
	{
		/*
		 * This can happen for obsolete inodes, i_dentry->d_inode of obsolete
		 * inodes holds reference to root inode of subfs.
		 * Also file inodes have hidden dentry without inode.
		 */
		struct dentry *tmp = inode->u.supermount_i.i_dentry;
		inode->u.supermount_i.i_dentry = 0;
		dput(tmp);
	}
	if (inode->i_count != 1)
		return;

	if (inode->u.supermount_i.i_hidden)
		supermount_debug("!!!!! WARNING !!!!! Hidden inode still present on %ld\n", inode->u.supermount_i.i_hidden->i_ino);

#if 0
	if (inode->u.supermount_i.i_hidden) {
		supermount_panic (inode->i_sb, "supermount_put_inode", 
				  "Hidden inode still present on "
				  "released inode");
	}
#endif
}

int supermount_permission (struct inode * inode, int mask)
{
	struct inode *subi;
	int rc = 0;
	int mode;
	
	supermount_dump_inode(inode);

	subi = subiget(inode);
	if (!subi)
	{
		supermount_go_inactive(inode->i_sb);
		return -EIO;
	}

	mode = subi->i_mode;
	supermount_debug("%o, %o, %d\n", mode, mask, capable(CAP_DAC_OVERRIDE));

	if (subi->i_op && subi->i_op->permission) {
/*		if (mask & MAY_WRITE)
			supermount_get_write_access(inode->i_sb);*/
		rc = subi->i_op->permission(subi, mask);
/*		if (mask & MAY_WRITE)
			supermount_put_write_access(inode->i_sb);*/
		goto out;
	}

	if ((mask & S_IWOTH) && IS_RDONLY(subi) &&
			(S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
	{
		rc = -EROFS; /* Nobody gets write access to a read-only fs */
		goto out;
	}
	if ((mask & S_IWOTH) && IS_IMMUTABLE(subi))
	{
		rc = -EACCES; /* Nobody gets write access to an immutable file */
		goto out;
	}
	if (current->fsuid == inode->i_uid)
		mode >>= 6;
	else if (in_group_p(inode->i_gid))
		mode >>= 3;

	if (((mode & mask & S_IRWXO) == mask) || capable(CAP_DAC_OVERRIDE))
		goto out;

	/* read and search access */
	if ((mask == S_IROTH) ||
			(S_ISDIR(mode)  && !(mask & ~(S_IROTH | S_IXOTH))))
		if (capable(CAP_DAC_READ_SEARCH))
			goto out;

	rc = -EACCES;
out:
	iput(subi);
	supermount_go_inactive(inode->i_sb);
	return rc;
}
