/*
 *  linux/fs/supermount/namei.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/namei.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/namei.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 * $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 */

/* This file handles almost all of the normal filesystem running of
   supermount.  We don't have to deal to much with the subfs
   interface, since we just pass subinodes out to the application on
   demand. */

#include <asm/segment.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>

/*
 * Find a new superinode, avoiding collisions with old
 * obsolete superinodes. 
 */
struct inode *get_possible_super_inode(struct super_block *sb, ino_t ino)
{
	unsigned long new_ino, incr=1237;
	struct inode *superi;

	new_ino = ino + incr;
	for ( ; (superi = iget(sb, new_ino)) && superi->i_count > 1 &&
		      (inode_is_obsolete(superi) ||
		       superi->u.supermount_i.i_subino != ino); ) {
		iput(superi);
		new_ino += incr;
		new_ino &= 0xffffff;
		while (new_ino < 2)
		{
			new_ino += incr;
			incr = (incr << 1) ^ incr;
		}
		incr = (incr << 1) ^ incr;
	}

	supermount_dump_inode(superi);

	return superi;
}

/*
 * Try to attach subinode to superinode.
 *
 * If subinode is already attached, return superinode. If hidden dentry is 
 * present on superinode (it should be the only case) dput subdentry (subd),
 * we do not need it because there already is one of superinode.
 */

struct inode * supermount_attach(struct super_block *sb, struct dentry *subd) 
{
	struct inode *subi = subd->d_inode;
	struct inode *superi = 0; /* to shut gcc up */
	
	supermount_debug ("inode = %ld\n", subi->i_ino);

	if (subi->i_shadow) {
		supermount_debug ("return existing shadow %ld, count %d/%d, %lx\n",
				  subi->i_shadow->i_ino,
				  subi->i_shadow->i_count, subi->i_count, 
				  (unsigned long)subi->i_shadow->u.supermount_i.i_dentry);
		if (S_ISDIR(subi->i_mode))
			subi->i_shadow->i_count++;
		else
			subi->i_count++;
		dput(subd);
		return subi->i_shadow;
	}

	/*
	 * Find a new superinode, avoiding collisions with old
	 * obsolete superinodes.
	 */

	superi = get_possible_super_inode(sb, subi->i_ino);

	/*
	 * We now have a new superinode which has been refcounted by
	 * one.  Next we have to attach the subinode and subdentry to it. 
	 */

	superi->u.supermount_i.i_subino = subi->i_ino;
	superi->u.supermount_i.i_dentry = subd;

	/*
	 * If we found a previously obsolete inode which is now no
	 * longer used, we can safely remove the obsolesence flag. 
	 */
	if (inode_is_obsolete(superi)) {
		superi->u.supermount_i.i_sb_version = 
			superi->i_sb->u.supermount_sb.s_version;
	}

/*
 *	superi->i_count == 1
 *	subi->i_count == 1
 */

	if (superi->u.supermount_i.i_hidden) {
		supermount_debug("!!!!!! WARNING !!!!!!, superinode has hidden inode, SHOULD NEVER HAPPEN\n");
		return superi;
	}

	/* 
	 * We have now done iget() on superi and subi.  Doing the
	 * attachment may incur another iget() on the subinode. 
	 */

	supermount_read_hidden_inode(superi);

	if (!superi->u.supermount_i.i_hidden) {
		supermount_debug ("!!!!! WARNING !!!!! read_hidden_inode didn't "
				  "attach inode");

		superi->u.supermount_i.i_dentry = NULL;
		iput(superi);
		return NULL;
	}

/*
 *	superi->i_count == 2
 *	subi->i_count == 2
 */
	/* 
	 * Reading the hidden inode refcounts the subinode
	 * again.  Dispose of that refcount now, since we don't
	 * need it --- we came into this routine already holding
	 * one.  
	 */
	iput(subi);
	if (!S_ISDIR(subi->i_mode))
	{
		list_del(&subd->d_alias);
		INIT_LIST_HEAD(&subd->d_alias);
		subd->d_inode = 0;
/*		d_instantiate(subd, superi);*/
		iput(superi);
	}

	return superi;
}

/*
 * Attach superinode to subinode and instantiate dentry with this inode
 * in case of directory and subinode in case of file.
 */

static void attach_and_prepare(struct super_block *sb, struct dentry *dentry,
							   struct dentry *subd, int create)
{
	struct inode *superi, *subi = subd->d_inode;

	supermount_dump_inode(sb->s_root->d_inode);
	supermount_debug("root dentry count: %d\n", sb->s_root->d_count);

	superi = supermount_attach(sb, subd);

	if (S_ISDIR(subi->i_mode))
	{
		if (create)
			d_instantiate(dentry, superi);
		else
			d_add(dentry, superi);
		dentry->d_op = &supermount_dir_dops;
	}
	else
	{
		if (create)
		{
			d_instantiate(dentry, subi);
			supermount_debug("dentry hashed: %d\n", !list_empty(&dentry->d_hash));
		}
		else
			d_add(dentry, subi);
		dentry->d_op = &supermount_file_dops;
		supermount_debug("subd count: %d\n", subd->d_count);
	}

	supermount_dump_inode(superi);
	supermount_debug("dentry count: %d/%d\n", dentry->d_count, (subd ? subd->d_count : -1));
	supermount_dump_inode(superi->i_sb->s_root->d_inode);
	supermount_debug("root dentry count: %d\n", superi->i_sb->s_root->d_count);
	supermount_debug("subroot dentry count: %d\n", superi->i_sb->u.supermount_sb.s_subfs->s_root->d_count);
}

static struct inode *get_subdir(struct inode *dir)
{
	struct inode *subdir;

	supermount_dump_inode(dir->i_sb->s_root->d_inode);
	supermount_debug("root dentry count: %d\n", dir->i_sb->s_root->d_count);
	
	if (!dir)
	{
		supermount_debug("!!!!! WARNING !!!!! dir is 0\n");
		return ERR_PTR(-EINVAL);
	}
	if (!S_ISDIR(dir->i_mode))
	{
		supermount_debug("!!!!! WARNING !!!!! dir is not dir\n");
		return ERR_PTR(-EINVAL);
	}

	supermount_dump_inode(dir);

	if (!(subdir = subiget(dir)))
	{
		/* This now becomes the *easy* case. :-) */
		return ERR_PTR(-ENOENT);
	}

	supermount_dump_inode(dir);
	supermount_dump_normal_inode(subdir);

	supermount_dump_inode(dir->i_sb->s_root->d_inode);
	supermount_debug("root dentry count: %d\n", dir->i_sb->s_root->d_count);

	return subdir;
}

/*
 * Create subdentry using dir's hidden dentry as parent, do subfs lookup
 * with created subdentry. Attach subdentry and it's inode (result of lookup)
 * to superinode.
 */

struct dentry *supermount_lookup (struct inode *dir, struct dentry *dentry)
{
	struct inode *subdir;
	struct dentry *err, *subdent;

	supermount_debug("lookup: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return (void *)subdir;

	subdent = d_alloc(dir->u.supermount_i.i_dentry, &dentry->d_name);
	if (!subdent)
	{
		iput(subdir);
		return ERR_PTR(-ENOMEM);
	}

	supermount_debug("lowlevel lookup: %s\n", subdent->d_name.name);

	err = subdir->i_op->lookup(subdir, subdent);
	d_drop(subdent); /* we don't need it on hash lists */
	/* Now we have subinode, attached to subdent. */

	iput(subdir); /* we don't need it anymore */

	if (err)
	{
		dput(subdent);
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %ld\n", (long)err);
		return err;
	}

	if (!subdent->d_inode)
	{
		/* not found */
		dput(subdent);
		d_add(dentry, NULL);
		return NULL;
	}

	supermount_dump_normal_inode(subdir);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 0);
	
/*	supermount_debug("hash_list_empty: %d\n", list_empty(&dentry->d_hash)); */

	supermount_go_inactive(dir->i_sb);
	if (S_ISDIR(dentry->d_inode->i_mode))
	{
		supermount_dump_inode(dentry->d_inode);
	}
	else
	{
		supermount_dump_inode(dentry->d_inode->i_shadow);
	}

	return 0;
}

int supermount_create (struct inode * dir, struct dentry *dentry, int mode)
{
	struct inode *subdir;
	struct dentry *subdent;
	int rc;

	supermount_debug("create: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->create) {
		iput(subdir);
		return -EPERM;
	}

	supermount_debug("dentry hashed: %d\n", !list_empty(&dentry->d_hash));

	subdent = d_alloc(dir->u.supermount_i.i_dentry, &dentry->d_name);
	if (!subdent)
	{
		iput(subdir);
		return -ENOMEM;
	}

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		dput(subdent);
		return rc;
	}

	supermount_debug("lowlevel create: %s\n", subdent->d_name.name);

	rc = subdir->i_op->create(subdir, subdent, mode);
	/* Now we have subinode, attached to subdent. */
	iput(subdir); /* We don't need it anymore*/

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);
	supermount_go_inactive(dir->i_sb);

	if (rc)
	{
		dput(subdent);
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		return rc;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		dput(subdent);
		d_instantiate(dentry, NULL);
		return 0;
	}

	supermount_dump_normal_inode(subdir);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
	
/*	supermount_debug("hash_list_empty: %d\n", list_empty(&dentry->d_hash));*/

	supermount_dump_inode(dentry->d_inode->i_shadow);
	supermount_debug("dentry %08lx\n", (unsigned long)dentry);

	return 0;
}

int supermount_unlink (struct inode *dir, struct dentry *dentry)
{
	struct inode *subdir;
	int rc;

	supermount_debug("unlink: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->unlink) {
		iput(subdir);
		return -EPERM;
	}

	if (!dentry->d_inode)
	{
		iput(subdir);
		return -ENOENT;
	}

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		return rc;
	}

	supermount_debug("lowlevel unlink: %s\n", dentry->d_name.name);

	rc = subdir->i_op->unlink(subdir, dentry);
	iput(subdir); /* We don't need it anymore*/

	return rc;
}

int supermount_mkdir (struct inode * dir, struct dentry *dentry, int mode)
{
	struct inode *subdir;
	struct dentry *subdent;
	int rc;

	supermount_debug("mkdir: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->mkdir) {
		iput(subdir);
		return -EPERM;
	}

	subdent = d_alloc(dir->u.supermount_i.i_dentry, &dentry->d_name);
	if (!subdent)
	{
		iput(subdir);
		return -ENOMEM;
	}

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		dput(subdent);
		return rc;
	}

	supermount_debug("lowlevel create: %s\n", subdent->d_name.name);

	rc = subdir->i_op->mkdir(subdir, subdent, mode);
	/* Now we have subinode, attached to subdent. */
	iput(subdir); /* We don't need it anymore*/

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);
	supermount_go_inactive(dir->i_sb);

	if (rc)
	{
		dput(subdent);
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		return rc;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		dput(subdent);
		d_instantiate(dentry, NULL);
		return 0;
	}

	supermount_dump_normal_inode(subdir);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
	
/*	supermount_debug("hash_list_empty: %d\n", list_empty(&dentry->d_hash));*/

	supermount_dump_inode(dentry->d_inode);
	supermount_debug("dentry %08lx\n", (unsigned long)dentry);

	return 0;
}

int supermount_rmdir (struct inode *dir, struct dentry *dentry)
{
	struct inode *subdir;
	struct dentry *subdent;
	int rc;

	supermount_debug("rmdir: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->rmdir) {
		iput(subdir);
		return -EPERM;
	}

	if (!dentry->d_inode || !dentry->d_inode->u.supermount_i.i_dentry)
	{
		iput(subdir);
		return -ENOENT;
	}

	subdent = dentry->d_inode->u.supermount_i.i_dentry;

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		return rc;
	}

	supermount_debug("lowlevel rmdir: %s\n", subdent->d_name.name);

	rc = subdir->i_op->rmdir(subdir, subdent);
	iput(subdir); /* We don't need it anymore*/
	/*	subdent will be dput by top level dput(dentry) */

	if (!rc)
		d_delete(dentry);       /* This also frees the inode */

	return rc;
}

int supermount_mknod (struct inode *dir, struct dentry *dentry, int mode, int dev)
{
	struct inode *subdir;
	struct dentry *subdent;
	int rc;

	supermount_debug("mknod: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->mknod) {
		iput(subdir);
		return -EPERM;
	}

	supermount_debug("dentry hashed: %d\n", !list_empty(&dentry->d_hash));

	subdent = d_alloc(dir->u.supermount_i.i_dentry, &dentry->d_name);
	if (!subdent)
	{
		iput(subdir);
		return -ENOMEM;
	}

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		dput(subdent);
		return rc;
	}

	supermount_debug("lowlevel create: %s\n", subdent->d_name.name);

	rc = subdir->i_op->mknod(subdir, subdent, mode, dev);
	/* Now we have subinode, attached to subdent. */
	iput(subdir); /* We don't need it anymore*/

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);
	supermount_go_inactive(dir->i_sb);

	if (rc)
	{
		dput(subdent);
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		return rc;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		dput(subdent);
		d_instantiate(dentry, NULL);
		return 0;
	}

	supermount_dump_normal_inode(subdir);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
	
/*	supermount_debug("hash_list_empty: %d\n", list_empty(&dentry->d_hash));*/

	supermount_dump_inode(dentry->d_inode->i_shadow);
	supermount_debug("dentry %08lx\n", (unsigned long)dentry);

	return 0;
}

int supermount_symlink (struct inode * dir, struct dentry *dentry,
	const char * symname)
{
	struct inode *subdir;
	struct dentry *subdent;
	int rc;

	supermount_debug("symlink: %s (parent: %08lx), directory: %ld\n",
		dentry->d_name.name, (unsigned long)dentry->d_parent,
		dir->i_ino);

	subdir = get_subdir(dir);
	if (IS_ERR(subdir))
		return PTR_ERR(subdir);

	if (!subdir->i_op || !subdir->i_op->symlink) {
		iput(subdir);
		return -EPERM;
	}

	supermount_debug("dentry hashed: %d\n", !list_empty(&dentry->d_hash));

	subdent = d_alloc(dir->u.supermount_i.i_dentry, &dentry->d_name);
	if (!subdent)
	{
		iput(subdir);
		return -ENOMEM;
	}

	rc = supermount_get_write_access(dir->i_sb);
	if (rc) {
		iput(subdir);
		dput(subdent);
		return rc;
	}

	supermount_debug("lowlevel symlink: %s->%s\n", symname, subdent->d_name.name);

	rc = subdir->i_op->symlink(subdir, subdent, symname);
	/* Now we have subinode, attached to subdent. */
	iput(subdir); /* We don't need it anymore*/

	supermount_debug("mode: %03o\n", subdent->d_inode->i_mode);
	mark_subfs_dirty(dir->i_sb);
	supermount_put_write_access(dir->i_sb);
	supermount_go_inactive(dir->i_sb);

	if (rc)
	{
		dput(subdent);
		supermount_debug("!!!!! WARNING !!!!! lowlevel error: %d\n", rc);
		return rc;
	}

	if (!subdent->d_inode)
	{
		supermount_debug("!!!!! WARNING !!!!! lowlevel didn't return inode\n");
		dput(subdent);
		d_instantiate(dentry, NULL);
		return 0;
	}

	supermount_dump_normal_inode(subdir);
	supermount_dump_normal_inode(subdent->d_inode);

	/* It worked, so now create a shadow supermount inode for the
           result... */
	attach_and_prepare(dir->i_sb, dentry, subdent, 1);
	
/*	supermount_debug("hash_list_empty: %d\n", list_empty(&dentry->d_hash));*/

	supermount_dump_inode(dentry->d_inode->i_shadow);
	supermount_debug("dentry %08lx\n", (unsigned long)dentry);

	return 0;
}

int supermount_rename (struct inode * olddir, struct dentry *olddentry,
						struct inode * newdir, struct dentry *newdentry)
{
	struct inode *oldsubdir, *newsubdir;
	struct dentry *oldsubdent, *newsubdent;
	int rc;

	supermount_debug("rename: %s(%08x)->%s(%08x) (parent: %08lx/%08lx), directory: %ld/%ld\n",
		olddentry->d_name.name, (unsigned int)olddentry->d_inode,
		newdentry->d_name.name, (unsigned int)newdentry->d_inode,
		(unsigned long)olddentry->d_parent, (unsigned long)newdentry->d_parent, 
		olddir->i_ino, newdir->i_ino);

	if (!olddentry->d_inode)
		return -EINVAL;
	if (olddentry->d_inode->i_shadow)
		oldsubdent = olddentry;
	else
		oldsubdent = olddentry->d_inode->u.supermount_i.i_dentry;
	if (!olddentry)
		return -EINVAL;

	oldsubdir = get_subdir(olddir);
	if (IS_ERR(oldsubdir))
		return PTR_ERR(oldsubdir);
	if (!oldsubdir->i_op || !oldsubdir->i_op->rename) {
		iput(oldsubdir);
		return -EPERM;
	}
	newsubdir = get_subdir(newdir);
	if (IS_ERR(newsubdir))
		return PTR_ERR(newsubdir);

	if (!newdentry->d_inode)
	{
		newsubdent = d_alloc(newdir->u.supermount_i.i_dentry, &newdentry->d_name);
		if (!newsubdent)
			rc = -ENOMEM;
	}
	else
	{
		if (newdentry->d_inode->i_shadow)
			newsubdent = dget(newdentry);
		else
			newsubdent = dget(newdentry->d_inode->u.supermount_i.i_dentry);
		if (!newsubdent)
			rc = -EINVAL;
	}
	if (!newsubdent)
	{
		iput(oldsubdir);
		iput(newsubdir);
		return rc;
	}

	rc = supermount_get_write_access(olddir->i_sb);
	if (rc) {
		iput(oldsubdir);
		iput(newsubdir);
		dput(newsubdent);
		return rc;
	}

	supermount_debug("lowlevel rename: %s->%s\n", oldsubdent->d_name.name,
												  newsubdent->d_name.name);

	rc = oldsubdir->i_op->rename(oldsubdir, oldsubdent, newsubdir, newsubdent);
	iput(oldsubdir); /* don't need it anymore */
	iput(newsubdir); /* don't need it anymore */

	supermount_put_write_access(olddir->i_sb);
	supermount_go_inactive(olddir->i_sb);

	dput(newsubdent);

	return rc;
}
