/*
 *  linux/fs/supermount/shadow.c
 *
 *  Original version:
 *      Copyright (C) 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 * $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 *
 *  Manage shadow inodes.
 * 
 *  Supermount maintains shadow inodes: supermount inodes which shadow
 *  inodes in the physically mounted subfilesystem.  The attachment of
 *  subinodes to their shadow superinodes is managed in this file.  We
 *  create the attachment on demand, and deattach the pair whenever the
 *  last reference to a subinode is released.  
 *
 *  The VFS also passes calls up to an inode's shadow when certain 
 *  interesting events occur, such incrementing or decrementing the write
 *  access count on a subinode.
 */

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>

static void supermount_shadow_release_inode (struct inode *);
static int  supermount_shadow_get_write_access (struct super_block *);
static void supermount_shadow_put_write_access (struct super_block *);
static int supermount_shadow_open_file(struct inode *inode, struct file *file);
static void supermount_shadow_release_file(struct inode *inode, struct file *file);
static int supermount_check_writeable(struct inode *inode);

/* 
 * Deal with the shadow interface.  The shadow inode ops are called by
 * the VFS whenever anything appropriately interesting happens to a
 * shadowed inode.  NOTE: The event has happened to the shadowed inode,
 * but it is the shadowing inode which gets the shadow callback.
 */

struct inode_shadow_operations supermount_shadow_iops = {
	supermount_shadow_open_file,
	NULL/*supermount_shadow_write*/,
	supermount_shadow_release_inode,
	supermount_shadow_release_file,
	NULL/*supermount_shadow_test_write_access*/,
	supermount_shadow_get_write_access,
	supermount_shadow_put_write_access,
	supermount_check_writeable
};

/* 
 * Take a given superinode and read an appropriate subinode for it.  The
 * i_ino number for the subinode is taken from the supermount i_subino
 * field if that is non-zero, and defaults to the inode number of the
 * superinode if it is zero. 
 *
 * If hidden dentry exist on superinode, it's inode is used as
 * subinode.
 *
 * The new subinode is attached to the superinode, and the refcount of
 * the superinode is incremented to reflect this.  The subinode's
 * refcount is also incremented: the function acts as an implicit iget()
 * on the subinode.
 */

void supermount_read_hidden_inode (struct inode * superi)
{
	struct super_block * subsb;
	struct inode * subi;

	supermount_dump_inode(superi);
	supermount_media_check(superi->i_sb);

	if (superi->i_sb->u.supermount_sb.s_state == SUPERMOUNT_UNMOUNTED) {
		supermount_debug ("!!!!! WARNING !!!!! Trying to read inode on unmounted media\n");
		return;
	}

	subsb = superi->i_sb->u.supermount_sb.s_subfs;

#if 0
	/* 
	 * If the superinode is obsolete, there's no point in trying to
	 * do any attachment. 
	 */
	if (inode_is_obsolete(superi)) {
		if (superi->u.supermount_i.i_subino)
			supermount_panic (superi->i_sb, "read_hidden_inode",
					  "Obsolete inode is "
					  "already attached");
		return;
	}
#endif
	if (superi->u.supermount_i.i_dentry)
	{
		superi->u.supermount_i.i_dentry->d_inode->i_count++;
		subi = superi->u.supermount_i.i_dentry->d_inode;
	}
	else
	{
		/*
		 * The root inode is a bit special.  For that one, the subinode
		 * is the root of the subfs, and we can't guarantee its inode
		 * number in advance.  Otherwise, just lookup the inode number
		 * we're given to find the appropriate subinode.
		 */
		if (superi->i_ino == SUPERMOUNT_ROOT_INO) {
			supermount_dump_inode(superi);
			subi = iget(subsb, subsb->s_root->d_inode->i_ino);
			supermount_dump_inode(superi);
			supermount_dump_normal_inode(subsb->s_root->d_inode);
		} else {
			if (!superi->u.supermount_i.i_subino)
				superi->u.supermount_i.i_subino = superi->i_ino;
			supermount_debug("get inode: %d\n", superi->u.supermount_i.i_subino);
			subi = iget(subsb, superi->u.supermount_i.i_subino);
			supermount_dump_normal_inode(subi);
		}
	}

	/*
	 * If the inode is already linked to its superinode, don't do
	 * any further processing.  Everything is fine.
	 */
	if (subi->i_shadow)
	{
		iput(subi);
		return;
	}

	subi->i_shadow = superi;
	superi->u.supermount_i.i_hidden = subi;
	superi->i_count++;

	superi->i_mode = subi->i_mode;
	superi->i_uid = subi->i_uid;
	superi->i_gid = subi->i_gid;
	superi->i_nlink = subi->i_nlink;
	superi->i_size = subi->i_size;
	superi->i_atime = subi->i_atime;
	superi->i_ctime = subi->i_ctime;
	superi->i_mtime = subi->i_mtime;
	superi->i_blksize = subi->i_blksize;
	superi->i_blocks = subi->i_blocks;
	superi->i_rdev = subi->i_rdev;
	superi->i_version = ++global_event;

	if (S_ISDIR(superi->i_mode))
		superi->i_op = &supermount_dir_iops;
	else
		superi->i_op = NULL;

	supermount_dump_inode(superi);
	supermount_dump_normal_inode(subi);
}

/*
 * Release link between sub- and superinode.
 */
void supermount_shadow_release_inode (struct inode *inode)
{
	supermount_dump_inode(inode);
	if (inode->u.supermount_i.i_hidden->i_count != 1)
		return;

	supermount_debug("subinode = %ld\n", 
				 inode->u.supermount_i.i_hidden->i_ino);

	inode->u.supermount_i.i_hidden->i_shadow = NULL;
	inode->u.supermount_i.i_hidden = NULL;
	iput(inode);
}

#if 0
static int supermount_shadow_test_write_access (struct inode *inode)
{
	supermount_dump_inode(inode);
	return supermount_test_write_access(inode->i_sb);
}
#endif

static int supermount_shadow_get_write_access (struct super_block *sb)
{
	int rc = 0;
//	supermount_dump_inode(inode);
	supermount_debug("supermount_shadow_get_write_access\n");
	rc = supermount_get_write_access(sb);
	if (!rc)
		supermount_iopen(sb);
	return rc;
}

static void supermount_shadow_put_write_access (struct super_block *sb)
{
	supermount_debug("supermount_shadow_put_write_access\n");
//	supermount_dump_inode(inode);
	mark_subfs_dirty(sb);
	supermount_put_write_access(sb);
	supermount_iclose(sb);
}

static int supermount_shadow_open_file(struct inode *inode, struct file *file)
{
	int rc = 0;

	supermount_dump_inode(inode);
	if (!rc)
		supermount_iopen(inode->i_sb);
	return rc;
}

static void supermount_shadow_release_file(struct inode *inode, struct file *file)
{
	if (S_ISDIR(inode->i_mode))
		return;
	supermount_dump_inode(inode);
	supermount_iclose(inode->i_sb);
}

static int supermount_check_writeable(struct inode *inode)
{
	supermount_dump_inode(inode);
	return !(inode->i_sb->s_flags & MS_RDONLY);
}
