/*
 *  linux/fs/supermount/super.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2. (C) 1999 Alexis Mikhailov
 *                                    (alexis@medinf.chuvashia.su)
 *
 * $Id: linux-2.2.17-supermount-0.1.5.chmou.patch,v 1.17 2000/11/25 18:21:13 devel Exp $
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>

#include <stdarg.h>

#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/smp_lock.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/supermount_fs.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/major.h>
#include <linux/cdrom.h>

#ifdef SUPERMOUNT_DEBUG
char supermount_debug_enable = 0;
#endif
char supermount_fail_statfs = 0;

static void umount_subfs(struct super_block *sb, int flags);

static struct super_operations supermount_sops = { 
	supermount_read_inode,
	supermount_write_inode,
	supermount_put_inode,		/* put_inode */
	NULL,		/* delete_inode */
	supermount_notify_change,
	supermount_put_super,
	supermount_write_super,
	supermount_statfs,
	NULL,				 /* supermount_remount */
};

static char error_buf[1024];

void supermount_error (struct super_block * sb, const char * function,
		       const char * fmt, ...)
{
	va_list args;

	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	printk (KERN_CRIT "SUPERMOUNT error (device %d/%d): %s: %s\n",
		MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
	/* @@ Do we want to do any special error handling here? */
}

NORET_TYPE void supermount_panic (struct super_block * sb,
				  const char * function,
				  const char * fmt, ...)
{
	va_list args;

	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	panic ("SUPERMOUNT panic (device %d/%d): %s: %s\n",
	       MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
}

void supermount_warning (struct super_block * sb, const char * function,
		   const char * fmt, ...)
{
	va_list args;

	va_start (args, fmt);
	vsprintf (error_buf, fmt, args);
	va_end (args);
	printk (KERN_WARNING "SUPERMOUNT-fs warning (device %d/%d): %s: %s\n",
		MAJOR(sb->s_dev), MINOR(sb->s_dev), function, error_buf);
}

/* Release the superblock and any resources it has reserved */
void supermount_put_super (struct super_block * sb)
{
	supermount_debug ("Dismounting superfs\n");
	
	sb->u.supermount_sb.s_dismount = 1;
	
	if (sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED)
		umount_subfs (sb, MNT_DO_NOT_LOCK_SEM);
	sb->s_dev = 0;

	MOD_DEC_USE_COUNT;
}

static int copy_option(const char **option, const char *val)
{
	char *tmp;
	supermount_debug ("assigning value \"%s\"\n", val);
	if (!val || !*val)
		return -EINVAL;
	tmp = (char *) kmalloc(1 + strlen(val), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;
	strcpy(tmp, val);
	*option = tmp;
	return 0;
}

static int parse_options (char * options, struct super_block *sb)
{
	char * this_char;
	char * value;
	int rc;
	
	if (!options)
		return 0;
	while ((this_char = options)) {
		if (!strncmp(this_char, "--,", 3))
			this_char += 2;
		if (*this_char == ',') {
			/* An empty option, or the option "--",
			   introduces options to be passed through to
			   the subfs */
			supermount_debug ("assigning remainder\n");
			return copy_option(&sb->u.supermount_sb.s_data,
					   ++this_char);
		}
		if ((options = strchr (this_char, ',')))
			*options++ = 0;
		
		if ((value = strchr (this_char, '=')) != NULL)
			*value++ = 0;

		supermount_debug ("parsing option \"%s\"\n", this_char);
		if (!strcmp (this_char, "fs")) {
			rc = copy_option(&sb->u.supermount_sb.s_type, value);
			if (rc) return rc;
		} else if (!strcmp (this_char, "dev")) {
			rc = copy_option(&sb->u.supermount_sb.s_devname, 
					 value);
			if (rc) return rc;
#ifdef SUPERMOUNT_DEBUG
		} else if (!strcmp (this_char, "debug")) {
			supermount_debug_enable = 1;
#endif
		} else if (!strcmp (this_char, "fail_statfs_until_mount")) {
			supermount_fail_statfs = 1;
		} else {
			printk ("supermount: "
				"Unrecognized mount option %s\n", this_char);
			return -EINVAL;
		}
	}
	return 0;
}

/* read_super: the main mount() entry point into the VFS layer. */
struct super_block * supermount_read_super (struct super_block * sb, 
										    void * data,
										    int silent)
{
	struct inode *inode;

	MOD_INC_USE_COUNT;
	supermount_debug("supermount read_super requested, flags: %ld\n", sb->s_flags);

	lock_super(sb);

	sb->s_blocksize = 1024;
	sb->s_blocksize_bits = 10;
	sb->s_magic = SUPERMOUNT_SUPER_MAGIC;
	sb->s_op = &supermount_sops;
	sb->u.supermount_sb.s_state = SUPERMOUNT_UNMOUNTED;
	sb->u.supermount_sb.s_default_mode = 0777 | S_IFDIR;
	sb->u.supermount_sb.s_mflags = sb->s_flags & MS_RMT_MASK;
	sb->u.supermount_sb.s_data = NULL;
	sb->u.supermount_sb.s_undermount = NULL;
	sb->u.supermount_sb.s_subfs = NULL;
	sb->u.supermount_sb.s_opencount = 0;
	sb->u.supermount_sb.s_writecount = 0;
	sb->u.supermount_sb.s_dirty = 0;
	sb->u.supermount_sb.s_rw = 0;
	sb->u.supermount_sb.s_dismount = 0;
	sb->u.supermount_sb.s_type = NULL;
	sb->u.supermount_sb.s_devname = NULL;

	if (parse_options ((char *) data, sb)) {
		sb->s_dev = 0;
		supermount_debug("super unlock super\n");
		unlock_super(sb);
		supermount_debug("super super unlocked\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	if (!sb->u.supermount_sb.s_type)
		copy_option(&sb->u.supermount_sb.s_type, "msdos");
	if (!sb->u.supermount_sb.s_devname)
		copy_option(&sb->u.supermount_sb.s_devname, "/dev/fd0");

	sb->s_flags = sb->u.supermount_sb.s_mflags;
	if (!(inode = iget(sb, SUPERMOUNT_ROOT_INO))) {
		supermount_error (sb, "supermount_read_super",
				  "get root inode failed\n");
		sb->s_dev = 0;
		supermount_debug("super unlock super\n");
		unlock_super(sb);
		supermount_debug("super super unlocked\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	
	supermount_dump_inode(inode);
	supermount_debug("umask: %06o\n", current->fs->umask);
	
	if (!(sb->s_root = d_alloc_root(inode, NULL))) {
		iput(inode);
		supermount_error (sb, "supermount_read_super",
				  "allocate root inode failed\n");
		sb->s_dev = 0;
		supermount_debug("super unlock super\n");
		unlock_super(sb);
		supermount_debug("super super unlocked\n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	sb->s_root->d_op = &supermount_dir_dops;

	supermount_debug("root dentry count: %d\n", sb->s_root->d_count);

	supermount_debug("super unlock super\n");
	unlock_super(sb);
	supermount_debug("super super unlocked\n");
	return sb;
}

void supermount_write_super (struct super_block * sb)
{
	supermount_debug("write_super\n");

	supermount_media_check(sb);
	if (sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED &&
	    sb->u.supermount_sb.s_rw) {
		struct super_block *tmp = sb->u.supermount_sb.s_subfs;
		if (tmp && tmp->s_op && tmp->s_op->write_super)
			tmp->s_op->write_super(tmp);
	}
	sb->s_dirt = 0;
}

int supermount_revalidate_inode(struct super_block *sb, struct inode *inode)
{
	if (supermount_media_check(sb) || (inode && inode_is_obsolete(inode)))
	{
		supermount_go_inactive(sb);
		supermount_debug(" - invalid\n");
		return 0;
	}
	supermount_go_inactive(sb);
	supermount_debug(" - valid\n");
	return 1;
}

/* 
 * When we get any attribute changes, we need to propogate them down
 * to the subinode.  
 */
int supermount_notify_change(struct dentry *dentry, struct iattr *attr)
{
	struct inode *subi, *inode = dentry->d_inode;
	struct dentry *subdent;
	int rc;

	if (!inode)
		return -EINVAL;

	if (inode->i_shadow)
	{
		subi = inode;
		subdent = dget(dentry);
		inode = subi->i_shadow;
		supermount_dump_inode (inode);
		supermount_iopen(inode->i_sb);
	}
	else
	{
		supermount_dump_inode (inode);

		subi = subiget (inode);
		if (!subi)
			return -ENOENT;
		supermount_iopen(inode->i_sb);

		subdent = dget(inode->u.supermount_i.i_dentry);
		if (!subdent) {
			iput(subi);
			supermount_iclose(inode->i_sb);
			return -ENOENT;
		}
		iput(subi);
	}

	rc = supermount_get_write_access(inode->i_sb);
	if (rc)	{
		dput(subdent);
		supermount_iclose(inode->i_sb);
		return rc;
	}

	if (subi->i_sb && subi->i_sb->s_op && subi->i_sb->s_op->notify_change) {
		rc = subi->i_sb->s_op->notify_change(subdent, attr);
	} else {
		rc = inode_change_ok(subi, attr);
		if (!rc)
			inode_setattr(subi, attr);
	}

	supermount_debug("rc: %d\n", rc);

	if (!rc) {
		/* 
		 * If it worked, then we need to mark the modification
		 * to the subfs, and we also need to propogate the
		 * change up to the shadowing inode.  
		 */
		mark_subfs_dirty(inode->i_sb);
		attr->ia_mode = subdent->d_inode->i_mode;
		attr->ia_uid = subdent->d_inode->i_uid;
		attr->ia_gid = subdent->d_inode->i_gid;
		attr->ia_size = subdent->d_inode->i_size;
		attr->ia_atime = subdent->d_inode->i_atime;
		attr->ia_mtime = subdent->d_inode->i_mtime;
		attr->ia_ctime = subdent->d_inode->i_ctime;
		attr->ia_valid = ATTR_UID | ATTR_GID | ATTR_MODE | ATTR_SIZE | 
						 ATTR_ATIME | ATTR_MTIME | ATTR_CTIME;
		supermount_debug("mode: %06o\n", attr->ia_mode);
		inode_setattr(inode, attr);
		supermount_debug("mode: %06o\n", inode->i_mode);
	}
	dput(subdent);

	supermount_put_write_access(inode->i_sb);
	supermount_iclose(inode->i_sb);
	supermount_go_inactive(inode->i_sb);

	return rc;
}

/* Check for media change, but without invalidating buffers and inodes */
static int just_check_disk_change(dev_t dev)
{
	if (!query_disk_change(dev))
		return 0;
	supermount_debug("Disk change detected on device %d/%d\n",
			 MAJOR(dev), MINOR(dev));
	return 1;
}

static void print_tree(struct dentry *root)
{
	struct list_head *head = root->d_subdirs.next;

	supermount_debug("%s, %d\n", root->d_name.name, root->d_count);
	if (root->d_inode != NULL)
		supermount_dump_normal_inode(root->d_inode);

	while (head != &root->d_subdirs)
	{
		struct dentry *dent = list_entry(head, struct dentry, d_child);

		print_tree(dent);
		head = head->next;
	}
}

static void release_hidden_dentries(struct dentry *root)
{
	struct list_head *head = root->d_subdirs.next;

	supermount_debug("%s, c(%d), %08lx\n", root->d_name.name, root->d_count,
		(unsigned long)root->d_inode);

	while (head != &root->d_subdirs)
	{
		struct dentry *dent = list_entry(head, struct dentry, d_child);

		release_hidden_dentries(dent);
		head = head->next;
	}

	supermount_debug("%s, c(%d), %08lx\n", root->d_name.name, root->d_count,
		(unsigned long)root->d_inode);

	if (root->d_inode != NULL && !root->d_inode->i_shadow /* directories */)
	{ 
		struct dentry *tmp = root->d_inode->u.supermount_i.i_dentry;

		supermount_dump_inode(root->d_inode);
		supermount_debug("hidden dentry %08x, count: %d\n", (unsigned int)tmp, (tmp ? tmp->d_count : -1));

		root->d_inode->u.supermount_i.i_dentry = NULL;
		if (tmp != NULL)
		{
			print_tree(tmp);
			supermount_debug("hidden dentry count: %d\n", (tmp ? tmp->d_count : -1));
			dput(tmp);
		}
	}
}

static void umount_subfs(struct super_block *sb, int flags)
{
	int retval;
	struct dentry *dent;
	
	supermount_debug("Trying to unmount device %s.\n",
			 sb->u.supermount_sb.s_devname);

	supermount_debug("Root: %08lx, Subroot: %08lx\n", (unsigned long)sb->s_root,
		(unsigned long)sb->u.supermount_sb.s_subfs->s_root);

	if (sb->s_root)
		supermount_debug("Root count: %d/%d\n", sb->s_root->d_count,
			sb->s_root->d_inode->i_count);
	if (sb->u.supermount_sb.s_subfs->s_root)
		supermount_debug("Subroot count: %d/%d\n", sb->u.supermount_sb.s_subfs->s_root->d_count,
			sb->u.supermount_sb.s_subfs->s_root->d_inode->i_count);

	/* Detach the subfs from the supermount root inode */
	dent = sb->s_root;
	if (dent)
	{
		/* We are unmounting because of media change */
		shrink_dcache_sb(sb);
		release_hidden_dentries(dent);
	}

	if (sb->u.supermount_sb.s_subfs->s_root->d_inode->i_count != 1)
		supermount_panic (sb, "umount_subfs",
				  "subfs root i_count = %d, should be 1\n",
				  sb->u.supermount_sb.s_subfs->s_root->d_inode->i_count);
	if (sb->u.supermount_sb.s_undermount->d_inode->i_count != 1)
		supermount_panic (sb, "umount_subfs",
				  "undermount i_count = %d, should be 1\n",
				  sb->u.supermount_sb.s_subfs->s_root->d_inode->i_count);
	sb->u.supermount_sb.s_state = SUPERMOUNT_UNMOUNTED;

	print_tree(sb->u.supermount_sb.s_subfs->s_root);

	supermount_debug("Undermount count: %d\n", sb->u.supermount_sb.s_undermount->d_count);
	retval = do_umounti(dget(sb->u.supermount_sb.s_subfs->s_root), flags);
	supermount_debug("Unmounted.\n");

	if (retval)
		supermount_panic (sb, "umount_subfs",
				  "Help - can't umount subfs!!!");

	supermount_debug("subfs is unmounted.\n");

	supermount_debug("Undermount count: %d/%d\n", sb->u.supermount_sb.s_undermount->d_count, sb->u.supermount_sb.s_undermount->d_inode->i_count);
	supermount_debug("Root: %08lx, Subroot: %08lx\n", (unsigned long)sb->s_root,
		(unsigned long)sb->u.supermount_sb.s_subfs->s_root);

	if (sb->s_root)
		supermount_debug("Root count: %d/%d\n", sb->s_root->d_count,
			sb->s_root->d_inode->i_count);
	if (sb->u.supermount_sb.s_subfs->s_root)
		supermount_debug("Subroot count: %d/%d\n", sb->u.supermount_sb.s_subfs->s_root->d_count,
			sb->u.supermount_sb.s_subfs->s_root->d_inode->i_count);


	/* It would be better to d_alloc_root in read_super, dput in put_super,
	and in check/unmount use only inodes */
	sb->u.supermount_sb.s_undermount->d_inode->i_sb = 0;
	dput(sb->u.supermount_sb.s_undermount);

	sb->u.supermount_sb.s_undermount = 0;
	sb->u.supermount_sb.s_version = ++global_event;
	sb->u.supermount_sb.s_subfs = 0;
	sb->s_flags = sb->u.supermount_sb.s_mflags;
	if (dent)
		dent->d_inode->i_mode = sb->u.supermount_sb.s_default_mode;
	supermount_iclose(sb);
}

/* Return 0 (OK) if medium is valid. */
int supermount_media_check(struct super_block * sb) 
{
	int retval;
	mm_segment_t fs;

	/* If we are dismounting, never return a success value. */
	if (sb->u.supermount_sb.s_dismount)
		return 1;
	
	/* If there are still files open, we never bother checking the
           medium. */
	if (sb->u.supermount_sb.s_subfs && subfs_is_active(sb)) {
		supermount_debug ("subfs is active. opencount: %d, undermount: %d\n",
			sb->u.supermount_sb.s_opencount,
			sb->u.supermount_sb.s_undermount->d_mounts->d_inode->i_count);
		return 0;
	}

	supermount_debug ("starting (%sactive now)\n",
			  (sb->u.supermount_sb.s_subfs &&
			   subfs_is_active(sb)) ?
			  "" : "not ");
	
	lock_super(sb);

	/* Are we checking for mount or unmount/remount? */
	if (sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED) {
		/* Already mounted --- check for disk change or
		   disk not present */
		int dev = sb->u.supermount_sb.s_subfs->s_dev;
		if (!just_check_disk_change(dev)) {
			unlock_super(sb);
			return 0;
		}
		
		/* We have a disk change!  Unmount the subfs */
		supermount_debug ("trying to unmount\n");
		umount_subfs(sb, 0);
		/* The call to just_check_disk_change may clear the media-
		   changed flag on the device, so we need to force a media
		   invalidation.  We don't want to do this before unmounting
		   the subfs, naturally! */
		invalidate_media(dev);
		/* Now we are unmounted; fall through to the next check. */
	}
	
	if (sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED) {
		unlock_super(sb);
		return 0;
	}
	
	/* OK, we're unmounted now --- can we remount?  Please? */
	if (!sb->u.supermount_sb.s_undermount) {
		struct inode *tmp = iget(sb, SUPERMOUNT_HIDDEN_INO);
		if (!tmp)
			supermount_panic (sb, "supermount_media_check",
					  "Can't get root hidden inode!");

		if (!(sb->u.supermount_sb.s_undermount = d_alloc_root(tmp, NULL))) {
			iput(tmp);
			supermount_panic (sb, "supermount_media_check",
					  "Can't get root hidden inode!");
		}
		
		supermount_debug("Undermount count: %d\n", sb->u.supermount_sb.s_undermount->d_count);
		supermount_dump_inode(tmp);
		supermount_dump_inode(sb->u.supermount_sb.s_undermount->d_inode);
	}

	sb->u.supermount_sb.s_version = ++global_event;
	sb->s_root->d_inode->u.supermount_i.i_sb_version = global_event;
	supermount_debug ("trying to mount\n");
	supermount_debug ("fs=%s, dev=%s, data=%s, flags=%08x\n",
			  sb->u.supermount_sb.s_type,
			  sb->u.supermount_sb.s_devname,
			  sb->u.supermount_sb.s_data,
			  sb->u.supermount_sb.s_mflags);

	supermount_debug("umask: %06o\n", current->fs->umask);
	sb->u.supermount_sb.s_rw = 0;
	fs = get_fs();
	set_fs(KERNEL_DS);
	sb->s_flags = sb->u.supermount_sb.s_mflags;
	retval = do_mounti(sb->u.supermount_sb.s_undermount,
			   "<supermount>",
			   sb->u.supermount_sb.s_devname,
			   sb->u.supermount_sb.s_type,
			   sb->s_flags | MS_MGC_VAL,
			   sb->u.supermount_sb.s_data);
	if (retval == -30)
	{
		supermount_debug("RW mount failed, try RO\n");
		retval = do_mounti(sb->u.supermount_sb.s_undermount,
				   "<supermount>",
				   sb->u.supermount_sb.s_devname,
				   sb->u.supermount_sb.s_type,
				   sb->s_flags | MS_MGC_VAL | MS_RDONLY,
				   sb->u.supermount_sb.s_data);
		sb->s_flags |= MS_RDONLY;
	}
	else if (retval == 0)
	{
		supermount_debug("Ok, filesystem is RW, now let's remount it RO\n");
		fsync_dev(sb->u.supermount_sb.s_undermount->d_mounts->d_sb->s_dev);
		retval = do_remount_sb (sb->u.supermount_sb.s_undermount->d_mounts->d_sb,
					(sb->s_flags & MS_RMT_MASK)
					| MS_RDONLY,
					NULL);
	}
	set_fs(fs);
	
	if (retval) {
		sb->s_flags = sb->u.supermount_sb.s_mflags;
		dput(sb->u.supermount_sb.s_undermount);
		sb->u.supermount_sb.s_undermount = 0;
		supermount_debug ("Mount failed, errno %d\n", -retval);
		unlock_super(sb);
		return 1;
	}

	/* Hey --- success!!! */
	supermount_debug("Undermount count: %d\n", sb->u.supermount_sb.s_undermount->d_count);

	sb->u.supermount_sb.s_subfs =
			sb->u.supermount_sb.s_undermount->d_mounts->d_sb;
	if (!sb->u.supermount_sb.s_subfs->s_root)
		supermount_panic (sb, "supermount_media_check", 
				  "Mounted subfs has no root inode!");
	supermount_dump_normal_inode(sb->u.supermount_sb.s_subfs->s_root->d_inode);
	sb->u.supermount_sb.s_state = SUPERMOUNT_SUSPENDED;
	sb->s_root->d_inode->u.supermount_i.i_dentry =
			dget(sb->u.supermount_sb.s_subfs->s_root);
	supermount_debug ("Mount succeeded!, flags: %ld\n", sb->u.supermount_sb.s_subfs->s_flags);
	supermount_iopen(sb);
	unlock_super(sb);

	return 0;
}

struct file_operations * get_blkfops(unsigned int major);

/* 
 * Try to lock the drive door.  This is not guaranteed to work on all
 * hardware, but we try any tricks we know of anyway.  
 */
void supermount_lock_door (struct super_block * sb, int lock)
{
	kdev_t dev;
	int i;
	struct file_operations * fops;

	dev = sb->u.supermount_sb.s_subfs->s_dev;
	i = MAJOR(dev);
	if (i >= MAX_BLKDEV || (fops = get_blkfops(i)) == NULL)
		return;
	if (fops->mediactl == NULL)
		return;
	supermount_debug("invoking mediactl\n");
	fops->mediactl(dev, lock ? MEDIA_LOCK : MEDIA_UNLOCK, 0);
}

/* 
 * Check the state of the supermount filesystem after it has been
 * released by some inode.  If it is no longer active, do the necessary
 * cleanup: fsync the disk, remount it readonly if appropriate, and
 * unlock the drive door if possible.
 */

void supermount_go_inactive(struct super_block *sb)
{
	supermount_debug("Checking state, currently %d\n",
			 sb->u.supermount_sb.s_state);
	if (sb->u.supermount_sb.s_state == SUPERMOUNT_UNMOUNTED)
		return;
	
	if (subfs_is_active(sb))
		return;

	supermount_debug("going offline.\n");
	sb->u.supermount_sb.s_state = SUPERMOUNT_SUSPENDED;

	/*
	 * Do an initial fsync_dev.  We do this before locking the 
	 * superblock to minimise the duration of the lock. 
	 */
	if (subfs_is_dirty(sb)) {
		supermount_debug("subfs is dirty, sync super fs\n");
		fsync_dev(sb->s_dev);
		mark_subfs_clean(sb);
	}

	/* 
	 * We need to recheck the state of the filesystem now.
	 * It may have become active again.  If not, we'll do
	 * some cleanup under the superblock lock to tidy things
	 * up.  */
	lock_super(sb);

	if (subfs_is_active(sb) || 
	    sb->u.supermount_sb.s_state != SUPERMOUNT_SUSPENDED) {
		unlock_super(sb);
		return;
	}
		
	/*
	 * First thing is, has the filesystem been made
	 * read-write since it became active?  If so, now is the
	 * time to retire it back to read-only state.
	 */
		
	if (sb->u.supermount_sb.s_writecount) {
		supermount_panic (sb, "supermount_go_inactive", 
				  "idle filesystem is write accessed");
	}
		
	if (sb->u.supermount_sb.s_rw) {
		int retval;
			
		/*
		 * Yup, it's read-write --- remount it.
		 */

		supermount_debug ("remounting readonly\n");
		supermount_debug("umask: %06o\n", current->fs->umask);

		fsync_dev(sb->u.supermount_sb.s_subfs->s_dev);
		retval = do_remount_sb (sb->u.supermount_sb.s_subfs,
					(sb->s_flags & MS_RMT_MASK)
					| MS_RDONLY,
					NULL);
		if (retval) 
			supermount_panic (sb, "supermount_go_inactive",
					  "Error %d remounting "
					  "subfs readonly", retval);

		sb->u.supermount_sb.s_rw = 0;
	}

	/*
	 * We can now afford to unlock the drive door. 
	 */
	supermount_debug("unlocking door.\n");
	supermount_lock_door(sb, 0);

	unlock_super(sb);
}

/*
 * Try to put the filesystem into writable mode for the duration of
 * an open(O_RDWR) or an inode operation requiring write access.  
 */

int supermount_get_write_access (struct super_block * sb)
{
	int retval = 0;
	
	supermount_debug ("wcount currently %d\n",
			  sb->u.supermount_sb.s_writecount);

	lock_super (sb);

	if (sb->u.supermount_sb.s_writecount == 0) {
		/*
		 * We may already have write access to the filesystem
		 */
		if (!sb->u.supermount_sb.s_rw) {
			if (sb->s_flags & MS_RDONLY)
				retval = -EROFS;
			else {
				supermount_debug ("remounting read/write\n");
				supermount_debug("umask: %06o\n", current->fs->umask);
				
				retval = do_remount_sb 
					(sb->u.supermount_sb.s_subfs,
					 sb->s_flags & MS_RMT_MASK,
					 NULL);
			}
		}
		if (!retval && !subfs_is_active(sb)) {
			supermount_debug ("going online.\n");
			sb->u.supermount_sb.s_state = SUPERMOUNT_ONLINE;
		}
	}

	if (!retval) {
		sb->u.supermount_sb.s_writecount++;
		sb->u.supermount_sb.s_rw = 1;
	}
	
	supermount_debug ("supermount_get_write_access, "
			  "wcount bumped to %d\n",
			  sb->u.supermount_sb.s_writecount);

	unlock_super (sb);
	return retval;
}

/*
 * Once we're done, release write access.
 */
void supermount_put_write_access (struct super_block * sb)
{
	supermount_debug ("wcount currently %d\n",
			  sb->u.supermount_sb.s_writecount);

	if (!sb->u.supermount_sb.s_writecount) {
		supermount_panic (sb, "supermount_put_write_access", 
				  "filesystem not write accessed");
		return;
	}
	
	if (!sb->u.supermount_sb.s_rw) {
		supermount_panic (sb, "supermount_put_write_access", 
				  "filesystem not read/write");
	}

	sb->u.supermount_sb.s_writecount--;
}

int supermount_statfs (struct super_block * sb, struct statfs * buf, 
			int bufsize)
{
	mm_segment_t fs;
	struct statfs tmp;
	int rc = 0;

	if (!supermount_fail_statfs || (supermount_fail_statfs &&
			sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED))
		supermount_media_check(sb);

	if (sb->u.supermount_sb.s_state != SUPERMOUNT_UNMOUNTED) {
		struct super_block * tmpsb = 
			sb->u.supermount_sb.s_undermount->d_mounts->d_sb;
		fs = get_fs();
		set_fs(KERNEL_DS);
		if (tmpsb->s_op && tmpsb->s_op->statfs)
			rc = tmpsb->s_op->statfs(tmpsb, &tmp, sizeof(tmp));
		set_fs(fs);
	} else {
		tmp.f_bsize = sb->s_blocksize;
		tmp.f_blocks = 0;
		tmp.f_bfree = 0;
		tmp.f_bavail = 0;
		tmp.f_files = 0;
		tmp.f_ffree = 0;
		tmp.f_namelen = 0;
	}
	tmp.f_type = SUPERMOUNT_SUPER_MAGIC;
	supermount_go_inactive(sb);
	if (!rc)
		return copy_to_user(buf, &tmp, bufsize) ? -EFAULT : 0;
	else
		return rc;
}

static struct file_system_type supermount_fs_type = {
	"supermount",
	0,
	supermount_read_super,
	NULL
};

__initfunc(int init_supermount_fs(void))
{
	return register_filesystem(&supermount_fs_type);
}

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
	return init_supermount_fs();
}
        
void cleanup_module(void)
{
	unregister_filesystem(&supermount_fs_type);
}
                
#endif
